/*

    Program: rtjpegplay - Play RTjpeg movies recorded with rtjpegrecord
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>

/* libbgrab include */
#include "bgrab.h"

/* xutils includes */
#include "xutil.h"

/* RTjpeg includes */
#include "RTjpeg.h"

/*
 Default color depth is 16bpp
*/

static int XDEPTH = 2;

/* YUV to RGB conversion prototype */

void (*yuv_conversion) (__u8 * buf, __u8 * rgb, int stride);

/* Playback counter */

static int FRAMECOUNT = (-1);

/* Main program */

#define MAX_FILENAME_LEN	256

int
main (int argc, char *argv[])
{
  int i;
  int magic;
  int WIDTH, HEIGHT;
  struct xwinbuffer xwin;
  int in_loop = 1;
  KeySym key;
  int button;
  int mousex, mousey;
  char inputfilename[MAX_FILENAME_LEN];
  unsigned long int tables[128];
  FILE *inputfile;
  int streamsize, itemsread;
  __s8 *stream;
  unsigned char *buffer;
  char *screenbuffer;
  int fps;
  double start_time, end_time, delay_time;
  unsigned long usleep_time;
  
  /* Message */
  fprintf (stderr,
	   "\nrtjpegplay - play RTjpeg movies\nA. Schiffler, aschiffler@home.com 05/00\n");

  /* Set default inputfilename */
  strcpy (inputfilename, "output.rtj");

  /* Parse command line parameters */
  for (i = 1; i < argc; i++)
    {
      char *arg;
      if (argv[i][0] != '-')
	break;
      arg = argv[i];
      arg++;
      if (*arg == '-')
	arg++;
      if (strncmp (arg, "frames", 6) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      FRAMECOUNT = atoi (argv[i]);
	    }
	}
      if (strncmp (arg, "input", 5) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      strncpy (inputfilename, argv[i], MAX_FILENAME_LEN);
	    }
	}
      else if (strncmp (arg, "help", 4) == 0)
	{
	  /* Print local help */
	  printf ("\nOutput settings:\n");
	  printf (" -input filename    inputfilename (default: %s)\n",
		  inputfilename);
	  printf (" -frames #          play only # frames\n");
	  /* Leave */
	  exit (0);
	}
    }

  /* Set YUV to RGB conversion routine */
  switch (XDEPTH)
    {
    case 1:
      yuv_conversion = RTjpeg_yuvrgb8;
      break;
    case 2:
      yuv_conversion = RTjpeg_yuvrgb16;
      break;
    case 4:
      yuv_conversion = RTjpeg_yuvrgb32;
      break;
    default:
      fprintf (stderr, "Unsupported display depth: %i\n", XDEPTH);
      exit (-1);
      break;
    }

  /* Open input file or assign stdin */
  if (strlen (inputfilename) == 0)
    {
      inputfile = stdin;
    }
  else
    {
      if ((inputfile = fopen (inputfilename, "r")) == NULL)
	{
	  perror ("Cannot open inputfile");
	  exit (-1);
	}
    }

  /* Read magic code */
  fread (&magic, sizeof (magic), 1, inputfile);
  if (magic != RTJPEG_MAGIC)
    {
      fprintf (stderr, "Not a RTjpeg file.\n");
      exit (-1);
    }

  /* Read size */
  fread (&WIDTH, sizeof (WIDTH), 1, inputfile);
  fread (&HEIGHT, sizeof (HEIGHT), 1, inputfile);

  /* Sanity check size */
  if ((WIDTH < 32) || (WIDTH > 924) || (HEIGHT < 32) || (HEIGHT > 756))
    {
      fprintf (stderr, "Bad size in RTjpeg file: %ix%i.\n", WIDTH, HEIGHT);
      exit (-1);
    }

  /* Read framerate */
  fread (&fps, sizeof (fps), 1, inputfile);

  /* Sanity check fps */
  if ((fps!=25) && (fps!=30)) 
    {
      fprintf (stderr, "Bad framerate in RTjpeg file: %i.\n", fps);
      exit (-1);
    }

  /* Read table */
  fread (tables, sizeof (tables), 1, inputfile);

  /* Initialize RTjpeg decompressor */
  RTjpeg_init_decompress (tables, WIDTH, HEIGHT);

  /* Allocate buffers */
  stream = malloc (WIDTH * HEIGHT + (WIDTH * HEIGHT) / 2);
  buffer = malloc (WIDTH * HEIGHT + (WIDTH * HEIGHT) / 2);

  /* Open window */
  if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, XDEPTH) != 0)
    exit (-1);

  /* View loop */
  while (in_loop)
    {

      /* Get the time */
      start_time=timeGet();
      
      /* Read chunk size and chunk from the stream */
      itemsread = fread (&streamsize, sizeof (streamsize), 1, inputfile);

      /* Check file end */
      if ((!feof (inputfile)) && (itemsread != 0) )
	{
	  /* Read the data */
	  itemsread = fread (stream, streamsize, 1, inputfile);

	  /* Check file end */
	  if (!feof (inputfile) && (itemsread != 0))
	    {

	      /* Decompress image stream */
	      RTjpeg_decompressYUV420 (stream, buffer);
	      /* Convert and copy image directly to the screenbuffer */
	      screenbuffer = (char *) xwin.shmimage->data;
	      yuv_conversion (buffer, screenbuffer, 0);
	      /* Write some text to screenbuffer */
	      drawstring (&xwin, 8, 8, "Q - quit.", 0xffffff);
	      /* Draw it */
	      sync_xwinbuffer (&xwin);
	      /* Poll keyboard */
	      if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey))
		{
		  switch (key)
		    {

		      /* Quit */

		    case XK_q:
		      in_loop = 0;
		      break;
		    }
		}


	      /* Update Framecounter */
	      if (FRAMECOUNT > 0)
		{
		  FRAMECOUNT--;
		}

	      /* Break out of loop */
	      if (FRAMECOUNT == 0)
		{
        	  fprintf (stderr,"Playback stopped.\n");
		  in_loop = 0;
		}

           /* Get the time again */
           end_time=timeGet();

           /* Calculate delay to play at target framerate */
           delay_time=(double)1000000.0*((double)1.0/(double)(fps)-(end_time-start_time));
           if (delay_time>0.0) {
            usleep_time=(unsigned long)delay_time;
            usleep(usleep_time);
           }
           
	    }
	  else
	    {
     	      fprintf (stderr,"Not enough data in stream.\n");
	      in_loop = 0;
	    }
	}
      else
	{
	  fprintf (stderr,"End of file reached.\n");
	  in_loop = 0;
	}

    }				/* View loop */


  /* Close file */
  fclose (inputfile);
  /* Close window */
  if (close_xwinbuffer (&xwin) != 0)
    {
      exit (-1);
    }

  /* Free memory */
  free (stream);
  free (buffer);
  exit (0);
}
