/*

    Program: rtjpegrecord - use RTjpeg to record live video
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

/* libbgrab includes */
#include "bgrab.h"

/* RTJpeg includes */
#include "RTjpeg.h"

/* 
 Default capture size
*/

static int WIDTH = 512;
static int HEIGHT = 256;

/*
 Default input format for RTjpeg is YCrCb422
 
 VIDEOFORMAT = color depth and format setup for input (framegrabber)  
*/

#define VIDEOFORMAT     FORMAT_YUV420P

/* Default videomode */

static int VIDEOMODE=VIDEOMODE_NTSC;

/* Number of frames */

static int FRAMECOUNT;

/* Compression control */

static unsigned char Q, M;


FILE *outputfile;
int in_loop;

/* Main program */

#define MAX_FILENAME_LEN	256

int
main (int argc, char *argv[])
{
  struct fgdevice fg;
  int i;
  char fg_filename[MAX_FILENAME_LEN];
  char outputfilename[MAX_FILENAME_LEN];
  void *newimage;
  int magic;
  unsigned long int tables[128];
  __s8 *stream;
  long original_size, compressed_size, total_input, total_output;  
  int framecounter;
  int fps;
  int region;
  int channel=6;
   
  /* Message */
  fprintf (stderr, "\nrtjpegrecord - record live video\n\n");

  /* Set default inpout/output filename */
  strcpy (fg_filename,"/dev/video");
  strcpy (outputfilename,"output.rtj");

  /* Set RJjpeg parameter defaults */
  Q=128;
  M=2;
  
  /* Reset framecount */
  FRAMECOUNT=-1;

  /* Parse command line parameters for local program */
  for (i = 1; i < argc; i++)
    {
      char *arg;
      if (argv[i][0] != '-')
	break;
      arg = argv[i];
      arg++;
      if (*arg == '-')
	arg++;
      if (strncmp (arg, "videodev", 8) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      strncpy (fg_filename, argv[i], MAX_FILENAME_LEN);
	    }
	}
      if (strncmp (arg, "output", 6) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      strncpy (outputfilename, argv[i], MAX_FILENAME_LEN);
	    }
	}
      if (strncmp (arg, "Q", 1) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      Q=(abs(atoi(argv[i])) % 256);
	    }
	}
      if (strncmp (arg, "frames", 6) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      FRAMECOUNT=atoi(argv[i]);
	    }
	}
      if (strncmp (arg, "M", 1) == 0)
	{
	  i++;
	  if (i <= argc)
	    {
	      M=(abs(atoi(argv[i])) % 256);
	    }
	}
      else if (strncmp (arg, "help", 4) == 0)
	{
	  /* Print local help */
	  printf ("\nDevice settings:\n");
	  printf (" -videodev dev      input device (default: /dev/video)\n");
	  printf ("\nCompression settings:\n");
	  printf (" -Q #               quality factor (default: %i)\n",Q);
	  printf (" -M #               motion factor (default: %i)\n",M);
	  printf ("\nOutput settings:\n");
	  printf (" -frames #          number of frames to record\n");
	  printf (" -output file       write stream to file (default: %s)\n",outputfilename);
	  /* Leave */
	  exit (0);
	}
    }

  /* Open framegrabber structure */
  if (fg_open_device (&fg, fg_filename)!=0) exit(-1); 	
  if (fg_print_info (&fg)!=0) exit(-1);

  /* --- This will likely have to changed for YOUR setup ---------- */
  /* .............................................................. */
 
  if (fg_set_channel (&fg, CHANNEL_TUNER, VIDEOMODE)!=0) exit(-1);
  region=REGION_NTSC_CABLE;
  if (fg_set_frequency (&fg, region, channel)!=0) exit(-1);
 
  /* .............................................................. */
 
  if (fg_start_grab_image(&fg, WIDTH, HEIGHT, VIDEOFORMAT)!=0) exit(-1);

  /* Determine frame rate from videomode */
  if (VIDEOMODE==VIDEOMODE_NTSC) {
   fps=30;
  } else {
   fps=25;
  } 
  
  /* Initialize RTjpeg routines and create buffers */
  stream=malloc(WIDTH*HEIGHT+WIDTH*HEIGHT/2);
  RTjpeg_init_compress(tables, WIDTH, HEIGHT, Q);
  RTjpeg_init_mcompress();
  
  /* Open output file or assign stdout */
  if (strlen(outputfilename)==0) {
   outputfile=stdout;
  } else {
   if ((outputfile=fopen(outputfilename,"w"))==NULL) {
    perror ("Cannot open outputfile");
    exit(-1);
   }
  }
     
  /* Write magic, size, fps and table to the file */
  magic=RTJPEG_MAGIC;
  fwrite (&magic, sizeof(magic),1,outputfile);
  fwrite (&WIDTH, sizeof(WIDTH),1,outputfile);
  fwrite (&HEIGHT, sizeof(HEIGHT),1,outputfile);
  fwrite (&fps, sizeof(fps),1,outputfile);
  fwrite (tables, sizeof(tables),1,outputfile);
  
  /* Setup size counters */
  total_input=0;
  total_output=0;
  original_size=WIDTH*HEIGHT*2;

  /* Recording loop */
  in_loop=1;
  framecounter=0;
  while (in_loop)
    {

      /* Update framebuffer with new image */
      if ( (newimage=fg_get_next_image(&fg)) != NULL)
	{
	   /* Count */
	   framecounter++;
	   
	   /* Compress image */ 
           compressed_size=RTjpeg_mcompressYUV420 (stream, newimage, M, M >> 1);
           
           /* Write size and data to file */
           fwrite (&compressed_size, sizeof(compressed_size), 1, outputfile);
           fwrite (stream, compressed_size, 1, outputfile);
           
           /* Update sums */
           total_input += original_size;
           total_output += compressed_size;
           
           /* Output compression ratio */
           fprintf (stderr, "Frames: %i  Input: %li K  Output: %li K  Ratio: %5.1f       \r",framecounter,total_input/1024,total_output/1024,(float)total_input/(float)total_output);

           /* Counter abort */
           if (FRAMECOUNT>0) {
            FRAMECOUNT--;
            if (FRAMECOUNT==0) {
             in_loop=0;
            }
           }
        }
    }				/* Record loop */

  /* Close outputfile */
  fclose (outputfile);

  /* Turn off grabbing */
  if (fg_stop_grab_image(&fg)!=0) exit(-1); 	
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(0); 
}
