/*

    Program: Motion trigger within patches through difference images
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 *
 * Requires XWindows running at 16bpp color depth.
 * Requires libbgrab-1.8 or greater
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bgrab.h"
#include "xutil.h"

#include "netsocket.h"

/* Screen resolution */

#define DEPTH 	1
#define XDEPTH  2

#define DEFAULT_FILENAME	"difftrigger.cnf"

/* Framegrabber setup */

static int brightness;
static int contrast;
static int color;

/* Image saving counter */

int capturecount=0;

/* Display toggle */

static int display_on=1;
static int colorcount=0;

/* Detection control */

static int differencelimit=10;  /* Difference must be > for detection*/
static int countlimit=2;        /* Number of detected pixels must be > for trigger */

/* Largest number of areas to test */
#define MAX_PATCH_NUM	100
  
/* Areas to test for difference changes */
struct tpatch {
		 int xmin, xmax;
		 int ymin, ymax;
		 int trigger;
                 int manualtrigger;
                 int updated;
		} patch[MAX_PATCH_NUM];
static int patchnum=0;

char blankline[81]="                                                                              ";

#define LINE_LEN	1024

/* Command line defaults */

static int WIDTH, HEIGHT, INPUTCHANNEL,
           PORT, VIDEOMODE, MEMORIZE, USEX;
static char VIDEODEV[LINE_LEN];

void set_defaults()
{
 WIDTH=640;
 HEIGHT=480;
 INPUTCHANNEL=1;
 PORT=1443;
 strcpy (VIDEODEV,"/dev/video");
 VIDEOMODE=VIDEOMODE_PAL;
 MEMORIZE=0;
 USEX=1;
}

/* Configuration file settings */

void SaveDefaultSettings()
{
 FILE *deffile;
 int i;
 
 deffile=fopen(DEFAULT_FILENAME,"w");

 fprintf(deffile,"# Framegrabber controls\n");
 fprintf(deffile,"contrast	%i\n",contrast);
 fprintf(deffile,"brightness	%i\n",brightness);
 fprintf(deffile,"color		%i\n",color);
 fprintf(deffile,"# Screen display On/Off\n");
 fprintf(deffile,"display	%i\n",display_on);
 fprintf(deffile,"# Detection control\n");
 fprintf(deffile,"dlimit	%i\n",differencelimit);
 fprintf(deffile,"climit	%i\n",countlimit);

 fprintf (deffile,"# Patches\n"); 
 for (i=0; i<patchnum; i++) {
  fprintf(deffile,">%i %i %i %i\n",
      patch[i].xmin,
      patch[i].ymin,
      patch[i].xmax,
      patch[i].ymax
      );
 }
 
 fclose(deffile);

 fprintf (stderr,"Settings saved ...\n");
}

void LoadDefaultSettings()
{
 FILE *deffile;
 char aline[LINE_LEN];
 char tag[LINE_LEN];
 int value;
 int xmin,ymin,xmax,ymax;
  
 /* Open file */
 if ((deffile=fopen(DEFAULT_FILENAME,"r"))!=NULL) {
 
  /* Scan file for tags/value pairs */
  /* Fetch a line */
  fgets(aline,LINE_LEN,deffile);  
  while (!feof(deffile)) {
   /* Test line we read */   if (aline[0]!='#') {
   /* Ignore line starting with a # */
    if (aline[0]=='>') {
    /* Read line starting with a > as patch areas */
     sscanf(&aline[1],"%i %i %i %i",&xmin,&ymin,&xmax,&ymax);
     if (xmin<0) { xmin=0; } else if (xmin>(WIDTH-1)) xmin=WIDTH-1;
     if (xmax<0) { xmax=0; } else if (xmax>(WIDTH-1)) xmax=WIDTH-1;
     if (ymin<0) { ymin=0; } else if (ymin>(HEIGHT-1)) ymin=HEIGHT-1;
     if (ymax<0) { ymax=0; } else if (ymax>(HEIGHT-1)) ymax=HEIGHT-1;
     /* Check patch size and storage area */
     if ((xmin<xmax) && (ymin<ymax) && (patchnum<MAX_PATCH_NUM)) {
      printf("Patch %i: (%i,%i)-(%i,%i)\n",patchnum,xmin,ymin,xmax,ymax);
      patch[patchnum].xmin=xmin;
      patch[patchnum].ymin=ymin;
      patch[patchnum].xmax=xmax;
      patch[patchnum].ymax=ymax;
      patch[patchnum].trigger=0;
      patch[patchnum].manualtrigger=0;
      patch[patchnum].updated=0;
      patchnum++;
     } 
    } else { 
     /* Scan tag-names and their value */
     sscanf(aline,"%s %i",tag,&value);
   
     printf ("Tag=%s Value=%i\n",tag,value);
     
     if (strncasecmp(aline,"contrast",8)==0) {
      contrast=value;
     }
     if (strncasecmp(aline,"brightness",8)==0) {
      brightness=value;
     }
     if (strncasecmp(aline,"color",5)==0) {
      color=value;
     }
     if (strncasecmp(aline,"display",7)==0) {
      display_on=value;
     }
     if (strncasecmp(aline,"dlimit",6)==0) {
      differencelimit=value;
     }
     if (strncasecmp(aline,"climit",6)==0) {
      countlimit=value;
     }
    } 
   }  
   /* Get next line */
   fgets(aline,LINE_LEN,deffile);  
  }
 
  /* Done - now close */
  fclose(deffile);

 } else {
  fprintf (stderr,"Cannot load %s.\n",DEFAULT_FILENAME);
 } 
}

/* -------------- Main routine --------------- */

int main (int argc, char *argv[])
{
 int i,x,y,framecount;
 struct fgdevice fg;
 struct xwinbuffer xwin;
 struct netthread nt;
 int not_first_time=0;
 int in_loop=1;
 KeySym key;
 int button=0;
 int mousex, mousey;
 int captureflag=0;
 char cfilename[LINE_LEN];
 char patchnumber[LINE_LEN];
 FILE *capturefile;
 unsigned char *oldframe;
 unsigned char *curposframe;
 unsigned char *curposoldframe;
 unsigned short *curposscreen=NULL;
 struct tpatch *curpatch;
 long count;
 int linecolor,numbercolor;
 int difference;
 int differencelimit=40;
 int trigger;

 fprintf (stderr,"difftrigger - patchwise motion detection\n");
 fprintf (stderr,"A. Schiffler, aschiffler@home.com, 03/00-05/00\n\n");
     
 /* Parse command line */
 set_defaults();
 for (i=1; i<argc; i++) {
  char	*arg;
  if (argv[i][0]!='-') break;
  arg = argv[i];
  arg++;
  if (*arg == '-') arg++;
  if (strncmp(arg,"width",5)==0) {
   i++; 
   if (i<=argc) {
    WIDTH = atoi(argv[i]);
   }
  } else if (strncmp(arg,"height",6)==0) {
   i++; 
   if (i<=argc) {
    HEIGHT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"port",4)==0) {
   i++; 
   if (i<=argc) {
    PORT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"memorize",8)==0) {
   i++; 
   MEMORIZE=1;
  } else if (strncmp(arg,"channel",7)==0) {
   i++; 
   if (i<=argc) {
    INPUTCHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"videodev",8)==0) {
   i++; 
   if (i<=argc) {
    strcpy (VIDEODEV,argv[i]);
   }
  } else if (strncmp(arg,"ntsc",4)==0) {
   VIDEOMODE=VIDEOMODE_NTSC;
  } else if (strncmp(arg,"secam",5)==0) {
   VIDEOMODE=VIDEOMODE_SECAM;
  } else if (strncmp(arg,"nox",3)==0) {
   USEX=0;
  } else {
   /* Command line help */
   printf ("Command line parameters:\n");
   printf ("\nVideo settings:\n");
   printf (" -width #           width of images (default: %i)\n",WIDTH);
   printf (" -height #          height of images (default: %i)\n",HEIGHT);
   printf (" -channel #         input channel (default: %i)\n",INPUTCHANNEL);
   printf (" -videodev txt      video device (default: %s)\n",VIDEODEV);
   printf (" -ntsc              use NTSC video (default: pal)\n");
   printf (" -secam             use SECAM video (default: pal)\n");
   printf ("\nNetwork settings:\n");
   printf (" -port #            listening port (default: %i)\n",PORT);
   printf ("\nTrigger settings:\n");
   printf (" -memorize          remember ON state until next read (default: off)\n");
   printf ("\nOutput settings:\n");
   printf (" -nox               turn off XWindows use (default: use XWindows)\n");

   exit(0);
  }
 }
 
 /* Message */
 printf ("Videodevice: %s, Channel: %i, Size: %ix%i.\n",VIDEODEV, INPUTCHANNEL, WIDTH, HEIGHT);
 /* Framegrabber initialization */
 if (fg_open_device (&fg, VIDEODEV)!=0) exit(-1); 	
 if (fg_print_info (&fg)!=0) exit(-1);
 if (fg_set_channel (&fg,INPUTCHANNEL,VIDEOMODE)!=0) exit(-1);

 /* Read settings from card */
 contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
 brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
 color=fg_get_setting(&fg,SETTING_COLOUR)-256;

 /* Load defaults */
 LoadDefaultSettings();
 
 /* Set settings */
 fg_set_setting(&fg,SETTING_CONTRAST,contrast);
 fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
 fg_set_setting(&fg,SETTING_COLOUR,color);

 if (USEX) {
  /* Setup xwindows */ 
  if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, XDEPTH)!=0) exit(-1);
 }

 /* Setup continous grabbing */
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_GREY)!=0) exit(-1);
 fg_set_fps_interval(&fg,100);
 
 /* Setup network */
 if (start_network_socket(&nt,PORT,BE_VERBOSE)!=0) exit(-1);

 /* Allocate memory for old frame */
 oldframe=(char *)malloc(WIDTH*HEIGHT*DEPTH);

 /* View/processing loop */
 framecount=0;
 while (in_loop) { 
  framecount++;
  
  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {

    if (USEX) {
     if (display_on) {
      /* Copy image to screenbuffer */
      curposframe=fg.current_image;
      curposscreen=(unsigned short *)xwin.shmimage->data;
      for (i=0; i<fg.image_size; i++) {
       *curposscreen = fg.y8_to_rgb565[(int)*curposframe];
       curposscreen++;
       curposframe++;
      }
     }
    }
    
    /* Image capture to file on user request */
    if (captureflag) {
     sprintf(cfilename,"capture%i.pgm",capturecount);
     printf ("Writing %s ...\n",cfilename);
     capturefile=fopen(cfilename,"w");
     /* Write headers */
     fprintf(capturefile,"P5\n# graydetect\n%i %i\n255\n",WIDTH,HEIGHT);
     /* Dump screen */
     fwrite ((void *)fg.current_image,WIDTH*HEIGHT,1,capturefile);
     fclose(capturefile);
     capturecount++;
     captureflag=0;
    } 

    /* Calculate absolute difference between frame and last frame */
    curpatch=patch;
    for (i=0; i<patchnum; i++) {
     /* Mean */
     curposframe=fg.current_image;
     curposframe += (curpatch->xmin+curpatch->ymin*WIDTH);
     curposoldframe=oldframe;
     curposoldframe += (curpatch->xmin+curpatch->ymin*WIDTH);
     if (USEX) {
      curposscreen=(unsigned short *)xwin.shmimage->data;
      curposscreen += (curpatch->xmin+curpatch->ymin*WIDTH);
     }
     count=0;
     for (y=curpatch->ymin; y<=curpatch->ymax; y++) {
      for (x=curpatch->xmin; x<=curpatch->xmax; x++) {
       /* Calculate difference */
       if (*curposframe>*curposoldframe) {
        difference=(int)((unsigned char)*curposframe-(unsigned char)*curposoldframe);
       } else {
        difference=(int)((unsigned char)*curposoldframe-(unsigned char)*curposframe);
       }
       if (USEX) {
        /* Show difference */
        *curposscreen=fg.y8_to_rgb565[difference];
        /* Check limit, show limit */
        if (difference>differencelimit) {
         *curposscreen=(unsigned short)(31 << 11);
         count++;
        }
        curposscreen++;
       }
       curposframe++; 
       curposoldframe++;
      }
      curposframe += (WIDTH-curpatch->xmax+curpatch->xmin-1);
      curposoldframe += (WIDTH-curpatch->xmax+curpatch->xmin-1);
      if (USEX) {
       curposscreen += (WIDTH-curpatch->xmax+curpatch->xmin-1);
      }
     }
     /* Work out trigger */
     trigger=(count>countlimit);
     /* work in manual trigger and store value */
     shared_data_lock(&nt);
     if (MEMORIZE) {
      curpatch->trigger=curpatch->trigger | trigger | curpatch->manualtrigger;
     } else {
      curpatch->trigger=trigger | curpatch->manualtrigger;
     }
     curpatch->manualtrigger=0;
     curpatch->updated=1; 
     shared_data_unlock(&nt);
     /* Next patch */
     curpatch++;
    }

    if (USEX) {
     if (display_on) {
      /* Set color */
      colorcount++;
      linecolor=0xffff;
      if (((colorcount >> 4) % 2) == 0) {
       numbercolor=0xffff;
      } else {
       numbercolor=0x0000;
      }
      /* Draw patch boundaries */
      curpatch=patch;
      for (i=0; i<patchnum; i++) {
       /* Get trigger state */
       shared_data_lock(&nt);
       trigger=curpatch->trigger; 
       shared_data_unlock(&nt);
       if (trigger) {
        hline(&xwin,curpatch->xmin,curpatch->xmax,curpatch->ymin,linecolor);
        hline(&xwin,curpatch->xmin,curpatch->xmax,curpatch->ymax,linecolor);
        vline(&xwin,curpatch->xmin,curpatch->ymin,curpatch->ymax,linecolor);
        vline(&xwin,curpatch->xmax,curpatch->ymin,curpatch->ymax,linecolor);
        drawline(&xwin,curpatch->xmin,curpatch->ymin,curpatch->xmax,curpatch->ymax,linecolor);
        drawline(&xwin,curpatch->xmin,curpatch->ymax,curpatch->xmax,curpatch->ymin,linecolor);
       } else {
        hline(&xwin,curpatch->xmin,curpatch->xmin+3,curpatch->ymin,linecolor);
        hline(&xwin,curpatch->xmin,curpatch->xmin+3,curpatch->ymax,linecolor);
        hline(&xwin,curpatch->xmax-3,curpatch->xmax,curpatch->ymin,linecolor);
        hline(&xwin,curpatch->xmax-3,curpatch->xmax,curpatch->ymax,linecolor);
        vline(&xwin,curpatch->xmin,curpatch->ymin,curpatch->ymin+3,linecolor);
        vline(&xwin,curpatch->xmax,curpatch->ymin,curpatch->ymin+3,linecolor);
        vline(&xwin,curpatch->xmin,curpatch->ymax-3,curpatch->ymax,linecolor);
        vline(&xwin,curpatch->xmax,curpatch->ymax-3,curpatch->ymax,linecolor);
       }
       sprintf(patchnumber,"%i",i+1);
       drawstring(&xwin,curpatch->xmin+3,curpatch->ymin+3,patchnumber,numbercolor);
       curpatch++;
      }
       
      /* Draw screen */
      sync_xwinbuffer(&xwin);
     }
    }
    
    /* Store frame for next iteration */
    memcpy(oldframe,fg.current_image,WIDTH*HEIGHT*DEPTH);
  }
  
  if (USEX) {
   /* Poll keyboard - process mouse right after grabbing */
   if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
    switch (key) {
    
    case XK_q:
     in_loop=0;
     break; 

    case XK_t:
     captureflag=1;
     break; 

    case XK_1:
     if (patchnum>0) {
      shared_data_lock(&nt);
      patch[0].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_2:
     if (patchnum>1) {
      shared_data_lock(&nt);
      patch[1].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_3:
     if (patchnum>2) {
      shared_data_lock(&nt);
      patch[2].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_4:
     if (patchnum>3) {
      shared_data_lock(&nt);
      patch[3].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_5:
     if (patchnum>4) {
      shared_data_lock(&nt);
      patch[4].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_6:
     if (patchnum>5) {
      shared_data_lock(&nt);
      patch[5].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 
   
    case XK_7:
     if (patchnum>6) {
      shared_data_lock(&nt);
      patch[6].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_8:
     if (patchnum>7) {
      shared_data_lock(&nt);
      patch[7].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_9:
     if (patchnum>8) {
      shared_data_lock(&nt);
      patch[8].manualtrigger=1; 
      shared_data_unlock(&nt);
     }
     break; 

    case XK_c:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
     fprintf (stderr,"Contrast set to %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 
   
    case XK_v:
     contrast=fg_get_setting(&fg,SETTING_CONTRAST)+256;
     fprintf (stderr,"Contrast set to %i\n",contrast);
     fg_set_setting(&fg,SETTING_CONTRAST,contrast);
     break; 
   
    case XK_b:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
     fprintf (stderr,"Brightness set to %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 
   
    case XK_n:
     brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)+256;
     fprintf (stderr,"Brightness set to %i\n",brightness);
     fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
     break; 

    case XK_i:
     color=fg_get_setting(&fg,SETTING_COLOUR)-256;
     fprintf (stderr,"Color set to %i\n",color);
     fg_set_setting(&fg,SETTING_COLOUR,color);
     break; 
   
    case XK_o:
     color=fg_get_setting(&fg,SETTING_COLOUR)+256;
     fprintf (stderr,"Color set to %i\n",color);
     fg_set_setting(&fg,SETTING_COLOUR,color);
     break; 
   
    case XK_s:
     SaveDefaultSettings();
     break; 

    case XK_d:
     display_on++;
     display_on &= 1;
     if (display_on==0) {
      bzero(xwin.shmimage->data,2*fg.image_size);
      /* Draw screen */
      sync_xwinbuffer(&xwin);
     }
     fprintf (stderr,"Display_on set to %i\n",display_on);
     break; 

    case XK_l:
     if (differencelimit<254) {
      differencelimit++;
      fprintf (stderr,"differencelimit set to %i\n",differencelimit);
     }
     break; 

    case XK_k:
     if (differencelimit>0) {
      differencelimit--;
      fprintf (stderr,"differencelimit set to %i\n",differencelimit);
     }
     break; 

    case XK_u:
     if (countlimit<1000) {
      countlimit++;
      fprintf (stderr,"countlimit set to %i\n",countlimit);
     }
     break; 

    case XK_y:
     if (countlimit>0) {
      countlimit--;
      fprintf (stderr,"countlimit set to %i\n",countlimit);
     }
     break; 

    }
   }

   /* Print position information on mouse-press */
   if (button) {
    printf ("Mouse is at X=%i Y=%i\n",mousex,mousey);
    button=0;
   } 
   
  } // USEX
   
  /* Print fps */
  if (fg.framecount==0) {
   if (not_first_time) {
    printf ("FPS=%7.1f\n",fg_get_fps(&fg));
   } else {
    not_first_time=1;
   } 
  }
 }           

 /* Shutdown network */
 if (stop_network_socket(&nt)!=0) exit(-1);

 if (USEX) {
  if (close_xwinbuffer(&xwin)!=0) exit(-1);
 }
 
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	

 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 exit(0);
}

int command_to_output(char *inbuffer, char* outbuffer, int outbufferlen)
{
 int target_selection;
 int target_reference;
 
 /* Read number */
 target_selection=0;
 sscanf (inbuffer,"%i",&target_selection);

#ifdef DEBUG
 fprintf(stderr,"Received: %s\n",inbuffer);
#endif
  
 /* Convert input command to output string */
 if ((target_selection>0) && (target_selection<=MAX_PATCH_NUM)) {
   target_reference=target_selection-1;
   snprintf (outbuffer,outbufferlen,"%i %i %i\n",
  					target_selection,
  					patch[target_reference].trigger,
  					patch[target_reference].updated
  					);
 patch[target_reference].updated=0;

 if (MEMORIZE) {
  patch[target_reference].trigger=0;
 }

#ifdef DEBUG
   fprintf(stderr,"Sending: %s",outbuffer);   
#endif

  return(0);
 }
 
 return(-1);
}
