/*

    Program: delaygrab - blockwise, controllable image delay
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |  
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 * NOTE: Requires XWindows running at 16bit depth 565 weight.
 *       Requires a lot of memory.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "bgrab.h"
#include "xutil.h"

#define WIDTH	640
#define HEIGHT	480
#define DEPTH 	2

#define BLOCKSIZE	4
#define DELAYMAPWIDTH   (WIDTH/BLOCKSIZE)
#define DELAYMAPHEIGHT	(HEIGHT/BLOCKSIZE)
#define DELAYMAPSIZE	(DELAYMAPHEIGHT*DELAYMAPWIDTH)

/* Reduce this for less memory consumption */
/* ....................................... */

#define QUEUEDEPTH	48

/* Globals */

int x,y,i,xyoff,v;
struct fgdevice fg;
struct xwinbuffer xwin;
int not_first_time=0;
int in_loop=1;
KeySym key;
int button;
int mousex,mousey;
char *imagequeue,*curqueue;
int curqueuenum;
int *delaymap=NULL,*curdelaymap;
char *curpos,*curimage;
int curposnum;
double d;

/* Calculate one of three possible "delaymaps" */

void CreateDelaymap(int mode)
{
 fprintf(stderr,"Delaymap changed to mode %i ...\n",mode);
 /* Create delay map array */
 if (delaymap==NULL) {
  delaymap=(int *)malloc(DELAYMAPSIZE*sizeof(int));
 }
 /* Fill values */
 curdelaymap=delaymap;
 for (y=0; y<DELAYMAPHEIGHT; y++) {
  for (x=0; x<DELAYMAPWIDTH; x++) {
   switch (mode) {
    case 1:
     /* Random delay with square distribution */
     d = (double)rand()/(double)RAND_MAX;
     *curdelaymap = (int)(d*d*16.0);
     break;
    case 2:
     /* Stripes of increasing delay outward from center */
     if (x<(DELAYMAPWIDTH/2)) {
      v=(DELAYMAPWIDTH/2)-x;
     } else if (x>(DELAYMAPWIDTH/2)) {
      v=x-(DELAYMAPWIDTH/2);
     } else {
      v=0;
     }
     *curdelaymap=v/2;
     break;
    case 3:
     /* Rings of increasing delay outward from center */
     v = (int)sqrt((double)((x-(DELAYMAPWIDTH/2))*(x-(DELAYMAPWIDTH/2))+(y-(DELAYMAPHEIGHT/2))*(y-(DELAYMAPHEIGHT/2))));
     *curdelaymap=v/2;
     break;
   } // switch
   /* Clip values */
   if (*curdelaymap<0) {
    *curdelaymap=0;
   } else if (*curdelaymap>(QUEUEDEPTH-1)) {
    *curdelaymap=(QUEUEDEPTH-1);
   }
   curdelaymap++;
  }
 }
}

/* ------------- */

int main (int argc, char *argv[])
{
 int framecount=0;

 fprintf (stderr,"delaygrab - libbgrab testprogram\n\n");

 /* Initialize randomizer */  
 srand(time(NULL)); 

 /* create initial Delaymap */
 CreateDelaymap(3);

 if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, DEPTH)!=0) exit(-1);
   
 if (argc>1) {
  if (fg_open_device (&fg, argv[1])!=0) exit(-1); 	
 } else {
  if (fg_open_device (&fg, "/dev/video")!=0) exit(-1); 	
 }
 
 if (fg_print_info (&fg)!=0) exit(-1);

 /* This line might require change depending on YOUR input */
 /* ...................................................... */
 if (fg_set_channel (&fg,CHANNEL_TUNER,VIDEOMODE_NTSC)!=0) exit(-1);
 
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_RGB565)!=0) exit(-1);

 fg_set_fps_interval(&fg,25);

 /* Create storage queue */ 
 if ((imagequeue=(char *)malloc(QUEUEDEPTH*fg.image_size))==NULL) {
  fprintf(stderr,"Not enough memory for a %i frame queue (%i bytes) - exiting.\n",QUEUEDEPTH,QUEUEDEPTH*fg.image_size);
  in_loop=0;
 } else {
  curqueue=imagequeue;
  curqueuenum=0;
 }

 /* View loop */
 while (in_loop) { 

  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {
   /* Count the frames */
   framecount++;

   /* Update queue pointer */
   if (curqueuenum==0) {
    curqueuenum=QUEUEDEPTH-1;
    curqueue = imagequeue;
    curqueue += (fg.image_size*(QUEUEDEPTH-1));
   } else {
    curqueuenum--;
    curqueue -= (fg.image_size);
   }
   
   /* Copy image to queue */
   memcpy((char *)curqueue,(char *)fg.current_image,fg.image_size);

   /* Copy image blockwise to screenbuffer */
   curdelaymap=delaymap;
   for (y=0; y<DELAYMAPHEIGHT; y++) {
    for (x=0; x<DELAYMAPWIDTH; x++) {
     curposnum=((curqueuenum + (*curdelaymap)) % QUEUEDEPTH);
     xyoff= x*(BLOCKSIZE*DEPTH) + y*(BLOCKSIZE*WIDTH*DEPTH);
     // source
     curpos=imagequeue;
     curpos += (fg.image_size*curposnum);
     curpos += xyoff;
     // target
     curimage = (char *)xwin.shmimage->data;
     curimage += xyoff;
     // copy block
     for (i=0; i<BLOCKSIZE; i++) {
      memcpy(curimage,curpos,(BLOCKSIZE*2));
      curpos += (WIDTH*DEPTH);
      curimage += (WIDTH*DEPTH);
     }
     curdelaymap++;
    }
   }

   /* Say something ... */
   drawstring (&xwin, 8,  8, "Press Q to quit", framecount);
   drawstring (&xwin, 8, 24, "Press 1, 2 and 3 to switch effects", framecount);
   
   /* Draw it */
   sync_xwinbuffer(&xwin);
  }

  /* Poll keyboard */
  if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
   switch (key) {
    case XK_q:
     in_loop=0;
     break; 
    case XK_1:
     CreateDelaymap(1);
     break; 
    case XK_2:
     CreateDelaymap(2);
     break; 
    case XK_3:
     CreateDelaymap(3);
     break; 
   }
  }
  
  /* Print fps */
  if (fg.framecount==0) {
   if (not_first_time) {
    printf ("FPS=%7.2f\n",fg_get_fps(&fg));
   } else {
    not_first_time=1; 
   } 
  } 
 }           

 if (close_xwinbuffer(&xwin)!=0) exit(-1);
 
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	

 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 free (delaymap);

 exit(0);
}
