/*

    Program: Greyscale measurements of regions in B/W images
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 *
 * Requires XWindows running at 16bpp color depth.
 * Requires libbgrab-1.8 or greater
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bgrab.h"
#include "xutil.h"

#include "netsocket.h"

/* Screen resolution */

#define DEPTH 	1
#define XDEPTH  2

#define DEFAULT_FILENAME	"greydetect.cnf"

/* Framegrabber setup */

static int brightness;
static int contrast;
static int color;

/* Image saving counter */

int capturecount=0;

/* Display toggle */

static int display_on=1;
static int colorcount=0;

/* Largest number of areas to test */
#define MAX_PATCH_NUM	100
  
/* Areas to test for difference changes */
struct tpatch {
		 int xmin, xmax;
		 int ymin, ymax;
                 float iarea;		// Inverse area of patch for calculations
                 float mean;		// Mean of Y values in patch
		 float stddev;		// Standard deviation of Y values in patch
                 int updated;
		} patch[MAX_PATCH_NUM];
static int patchnum=0;

char blankline[81]="                                                                              ";

#define LINE_LEN	1024

/* Command line defaults */

static int WIDTH, HEIGHT, INPUTCHANNEL,
           PORT, VIDEOMODE, USEX;
static char VIDEODEV[LINE_LEN];

void set_defaults()
{
 WIDTH=640;
 HEIGHT=480;
 INPUTCHANNEL=1;
 PORT=1442;
 strcpy (VIDEODEV,"/dev/video");
 VIDEOMODE=VIDEOMODE_PAL;
 USEX=1;
}

/* Configuration file settings */

void SaveDefaultSettings()
{
 FILE *deffile;
 int i;
 
 deffile=fopen(DEFAULT_FILENAME,"w");

 fprintf(deffile,"# Framegrabber controls\n");
 fprintf(deffile,"contrast	%i\n",contrast);
 fprintf(deffile,"brightness	%i\n",brightness);
 fprintf(deffile,"color		%i\n",color);
 fprintf(deffile,"# Screen display On/Off\n");
 fprintf(deffile,"display	%i\n",display_on);

 fprintf (deffile,"# Patches\n"); 
 for (i=0; i<patchnum; i++) {
  fprintf(deffile,">%i %i %i %i\n",
      patch[i].xmin,
      patch[i].ymin,
      patch[i].xmax,
      patch[i].ymax
      );
 }
 
 fclose(deffile);

 fprintf(stderr,"Settings saved ...\n");
}

void LoadDefaultSettings()
{
 FILE *deffile;
 char aline[LINE_LEN];
 char tag[LINE_LEN];
 int value;
 int xmin,ymin,xmax,ymax;
  
 /* Open file */
 if ((deffile=fopen(DEFAULT_FILENAME,"r"))!=NULL) {
 
  /* Scan file for tags/value pairs */
  /* Fetch a line */
  fgets(aline,LINE_LEN,deffile);  
  while (!feof(deffile)) {
   /* Test line we read */   if (aline[0]!='#') {
   /* Ignore line starting with a # */
    if (aline[0]=='>') {
    /* Read line starting with a > as patch areas */
     sscanf(&aline[1],"%i %i %i %i",&xmin,&ymin,&xmax,&ymax);
     if (xmin<0) { xmin=0; } else if (xmin>(WIDTH-1)) xmin=WIDTH-1;
     if (xmax<0) { xmax=0; } else if (xmax>(WIDTH-1)) xmax=WIDTH-1;
     if (ymin<0) { ymin=0; } else if (ymin>(HEIGHT-1)) ymin=HEIGHT-1;
     if (ymax<0) { ymax=0; } else if (ymax>(HEIGHT-1)) ymax=HEIGHT-1;
     /* Check patch size and storage area */
     if ((xmin<xmax) && (ymin<ymax) && (patchnum<MAX_PATCH_NUM)) {
      printf("Patch %i: (%i,%i)-(%i,%i)\n",patchnum,xmin,ymin,xmax,ymax);
      patch[patchnum].xmin=xmin;
      patch[patchnum].ymin=ymin;
      patch[patchnum].xmax=xmax;
      patch[patchnum].ymax=ymax;
      patch[patchnum].iarea=1.0/(float)((xmax-xmin+1)*(ymax-ymin+1));
      patch[patchnum].mean=0.0;
      patch[patchnum].stddev=0.0;
      patch[patchnum].updated=0;
      patchnum++;
     } 
    } else { 
     /* Scan tag-names and their value */
     sscanf(aline,"%s %i",tag,&value);
   
     printf ("Tag=%s Value=%i\n",tag,value);
     
     if (strncasecmp(aline,"contrast",8)==0) {
      contrast=value;
     }
     if (strncasecmp(aline,"brightness",8)==0) {
      brightness=value;
     }
     if (strncasecmp(aline,"color",5)==0) {
      color=value;
     }
     if (strncasecmp(aline,"display",7)==0) {
      display_on=value;
     }
    } 
   }  
   /* Get next line */
   fgets(aline,LINE_LEN,deffile);  
  }
 
  /* Done - now close */
  fclose(deffile);

 } else {
  fprintf (stderr,"Cannot load %s.\n",DEFAULT_FILENAME);
 } 
}

/* -------------- Main routine --------------- */

int main (int argc, char *argv[])
{
 int i,x,y,framecount;
 struct fgdevice fg;
 struct xwinbuffer xwin;
 struct netthread nt;
 int not_first_time=0;
 int in_loop=1;
 KeySym key;
 int button=0;
 int mousex, mousey;
 int captureflag=0;
 char cfilename[LINE_LEN];
 char patchnumber[LINE_LEN];
 FILE *capturefile;
 unsigned char *curposframe;
 unsigned short *curposscreen;
 struct tpatch *curpatch;
 long count,count2;
 int linecolor,numbercolor;
 
 fprintf (stderr,"graydetect - patchwise grayscale measurements\n");
 fprintf (stderr,"A. Schiffler, aschiffler@home.com, 01/00-05/00\n\n");
     
 /* Parse command line */
 set_defaults();
 for (i=1; i<argc; i++) {
  char	*arg;
  if (argv[i][0]!='-') break;
  arg = argv[i];
  arg++;
  if (*arg == '-') arg++;
  if (strncmp(arg,"width",5)==0) {
   i++; 
   if (i<=argc) {
    WIDTH = atoi(argv[i]);
   }
  } else if (strncmp(arg,"height",6)==0) {
   i++; 
   if (i<=argc) {
    HEIGHT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"port",4)==0) {
   i++; 
   if (i<=argc) {
    PORT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"channel",6)==0) {
   i++; 
   if (i<=argc) {
    INPUTCHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"videodev",8)==0) {
   i++; 
   if (i<=argc) {
    strcpy (VIDEODEV,argv[i]);
   }
  } else if (strncmp(arg,"ntsc",4)==0) {
   VIDEOMODE=VIDEOMODE_NTSC;
  } else if (strncmp(arg,"secam",5)==0) {
   VIDEOMODE=VIDEOMODE_SECAM;
  } else if (strncmp(arg,"nox",3)==0) {
   USEX=0;
  } else {
   /* Command line help */
   printf ("Command line parameters:\n");
   printf ("\nVideo settings:\n");
   printf (" -width #           width of images (default: %i)\n",WIDTH);
   printf (" -height #          height of images (default: %i)\n",HEIGHT);
   printf (" -channel #         input channel (default: %i)\n",INPUTCHANNEL);
   printf (" -videodev txt      video device (default: %s)\n",VIDEODEV);
   printf (" -ntsc              use NTSC video (default: pal)\n");
   printf (" -secam             use SECAM video (default: pal)\n");
   printf ("\nNetwork settings:\n");
   printf (" -port #            listening port (default: %i)\n",PORT);
   printf ("\nOutput settings:\n");
   printf (" -nox               turn off XWindows use (default: use XWindows)\n");
   exit(0);
  }
 }
 
 /* Message */
 printf ("Videodevice: %s, Channel: %i, Size: %ix%i.\n",VIDEODEV, INPUTCHANNEL, WIDTH, HEIGHT);
 /* Framegrabber initialization */
 if (fg_open_device (&fg, VIDEODEV)!=0) exit(-1); 	
 if (fg_print_info (&fg)!=0) exit(-1);
 if (fg_set_channel (&fg,INPUTCHANNEL,VIDEOMODE)!=0) exit(-1);

 /* Read settings from card */
 contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
 brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
 color=fg_get_setting(&fg,SETTING_COLOUR)-256;

 /* Load defaults */
 LoadDefaultSettings();
 
 /* Set settings */
 fg_set_setting(&fg,SETTING_CONTRAST,contrast);
 fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
 fg_set_setting(&fg,SETTING_COLOUR,color);

 if (USEX) {
  /* Setup xwindows */ 
  if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, XDEPTH)!=0) exit(-1);
 }
 
 /* Setup continous grabbing */
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_GREY)!=0) exit(-1);
 fg_set_fps_interval(&fg,100);
 
 /* Setup network */
 if (start_network_socket(&nt,PORT,BE_VERBOSE)!=0) exit(-1);

 /* Clear textscreen */
 printf ("\033[H");
 for (i=0; i<40; i++) printf ("%s\n",blankline);

 /* View/processing loop */
 framecount=0;
 while (in_loop) { 
  framecount++;
  
  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {

    /* Back to top */
    printf ("\033[H");  
    /* Print header again */
    printf ("graydetect - patchwise grayscale measurements\n");
    printf ("A. Schiffler, ZKM, 01/00\n\n");

    if (USEX) {
     if (display_on) {
      /* Copy image to screenbuffer */
      curposframe=fg.current_image;
      curposscreen=(unsigned short *)xwin.shmimage->data;
      for (i=0; i<fg.image_size; i++) {
       *curposscreen = fg.y8_to_rgb565[(int)*curposframe];
       curposscreen++;
       curposframe++;
      }
     }
    }
    
    /* Image capture to file on user request */
    if (captureflag) {
     sprintf(cfilename,"capture%i.pgm",capturecount);
     printf ("Writing %s ...\n",cfilename);
     capturefile=fopen(cfilename,"w");
     /* Write headers */
     fprintf(capturefile,"P5\n# graydetect\n%i %i\n255\n",WIDTH,HEIGHT);
     /* Dump screen */
     fwrite ((void *)fg.current_image,WIDTH*HEIGHT,1,capturefile);
     fclose(capturefile);
     capturecount++;
     captureflag=0;
    } 

    /* Calculate patch statistical values */
    printf ("Greyscale measurements:\n");
    curpatch=patch;
    for (i=0; i<patchnum; i++) {
     /* Mean */
     curposframe=fg.current_image;
     curposframe += (curpatch->xmin+curpatch->ymin*WIDTH);
     count=0;
     count2=0;
     for (y=curpatch->ymin; y<=curpatch->ymax; y++) {
      for (x=curpatch->xmin; x<=curpatch->xmax; x++) {
       /* Convert to Y "in-place" */
       count += (long)*curposframe;
       count2 += ((long)*curposframe*(long)*curposframe);
       curposframe++; 
      }
      curposframe += (WIDTH-curpatch->xmax+curpatch->xmin-1);
     }
     /* Calc and store values */
     shared_data_lock(&nt);
     curpatch->mean = (float)count*curpatch->iarea;
     curpatch->stddev = (float)count2*curpatch->iarea - curpatch->mean*curpatch->mean;
     if (curpatch->stddev<=0.0) {
      curpatch->stddev=0.0;
     } else { 
      curpatch->stddev=(float)sqrt((double)curpatch->stddev);
     }
     /* Output patch data to screen */
     printf("Patch %i  (%3i,%3i)-(%3i,%3i) :  Mean =  %6.1f   Std. Deviation = %6.1f\n",i,curpatch->xmin,curpatch->ymin,curpatch->xmax,curpatch->ymax,curpatch->mean,curpatch->stddev);
     curpatch->updated = 1;
     shared_data_unlock(&nt);
     /* Next patch */
     curpatch++;
    }

    if (USEX) {
     if (display_on) {
      /* Set color */
      colorcount++;
      linecolor=0xffff;
      if (((colorcount >> 4) % 2) == 0) {
       numbercolor=0xffff;
      } else {
       numbercolor=0x0000;
      }
      /* Draw patch boundaries */
      curpatch=patch;
      for (i=0; i<patchnum; i++) {
       hline(&xwin,curpatch->xmin,curpatch->xmin+3,curpatch->ymin,linecolor);
       hline(&xwin,curpatch->xmin,curpatch->xmin+3,curpatch->ymax,linecolor);
       hline(&xwin,curpatch->xmax-3,curpatch->xmax,curpatch->ymin,linecolor);
       hline(&xwin,curpatch->xmax-3,curpatch->xmax,curpatch->ymax,linecolor);
       vline(&xwin,curpatch->xmin,curpatch->ymin,curpatch->ymin+3,linecolor);
       vline(&xwin,curpatch->xmax,curpatch->ymin,curpatch->ymin+3,linecolor);
       vline(&xwin,curpatch->xmin,curpatch->ymax-3,curpatch->ymax,linecolor);
       vline(&xwin,curpatch->xmax,curpatch->ymax-3,curpatch->ymax,linecolor);
       sprintf(patchnumber,"%i",i);
       drawstring(&xwin,curpatch->xmin+3,curpatch->ymin+3,patchnumber,numbercolor);
       curpatch++;
      }
       
      /* Draw screen */
      sync_xwinbuffer(&xwin);
     } 
    }
  }

  if (USEX) {
   /* Poll keyboard - process mouse right after grabbing */
   if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {
    switch (key) {
    
     case XK_q:
      in_loop=0;
      break; 

     case XK_t:
      captureflag=1;
      break; 
   
     case XK_c:
      contrast=fg_get_setting(&fg,SETTING_CONTRAST)-256;
      fprintf (stderr,"Contrast set to %i\n",contrast);
      fg_set_setting(&fg,SETTING_CONTRAST,contrast);
      break; 
   
     case XK_v:
      contrast=fg_get_setting(&fg,SETTING_CONTRAST)+256;
      fprintf (stderr,"Contrast set to %i\n",contrast);
      fg_set_setting(&fg,SETTING_CONTRAST,contrast);
      break; 
   
     case XK_b:
      brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)-256;
      fprintf (stderr,"Brightness set to %i\n",brightness);
      fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
      break; 
   
     case XK_n:
      brightness=fg_get_setting(&fg,SETTING_BRIGHTNESS)+256;
      fprintf (stderr,"Brightness set to %i\n",brightness);
      fg_set_setting(&fg,SETTING_BRIGHTNESS,brightness);
      break; 
 
     case XK_i:
      color=fg_get_setting(&fg,SETTING_COLOUR)-256;
      fprintf (stderr,"Color set to %i\n",color);
      fg_set_setting(&fg,SETTING_COLOUR,color);
      break; 
   
     case XK_o:
      color=fg_get_setting(&fg,SETTING_COLOUR)+256;
      fprintf (stderr,"Color set to %i\n",color);
      fg_set_setting(&fg,SETTING_COLOUR,color);
      break; 
   
     case XK_s:
      SaveDefaultSettings();
      break; 

     case XK_d:
      display_on++;
      display_on &= 1;
      if (display_on==0) {
       bzero(xwin.shmimage->data,2*fg.image_size);
       /* Draw screen */
       sync_xwinbuffer(&xwin);
      }
      fprintf (stderr,"Display_on set to %i\n",display_on);
      break; 
    }
   }
 
   /* Print position information on mouse-press */
   printf ("\nMouse position at click: ");
   if (button) {
    printf ("X=%i Y=%i\n",mousex,mousey);
    button=0;
   } else {
    printf ("\n");
   }
  }
   
  /* Print fps */
  printf ("Display speed: ");   
  if (fg.framecount==0) {
   if (not_first_time) {
    printf ("FPS=%7.1f\n",fg_get_fps(&fg));
   } else {
    not_first_time=1;
   } 
  } else {
   printf("\n");
  } 
 }           

 /* Shutdown network */
 if (stop_network_socket(&nt)!=0) exit(-1);

 if (USEX) {
  if (close_xwinbuffer(&xwin)!=0) exit(-1);
 }
  
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	

 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 exit(0);
}

int command_to_output(char *inbuffer, char* outbuffer, int outbufferlen)
{
 int target_selection;
 int target_reference;
 
 /* Read number */
 target_selection=0;
 sscanf (inbuffer,"%i",&target_selection);

#ifdef DEBUG
 fprintf(stderr,"Received: %s\n",inbuffer);
#endif
  
 /* Convert input command to output string */
 if ((target_selection>0) && (target_selection<=MAX_PATCH_NUM)) {
   target_reference=target_selection-1;
   snprintf (outbuffer,outbufferlen,"%i %7.3f %7.3f\n",
  					target_selection,
  					patch[target_reference].mean,
  					patch[target_reference].stddev
  					);
  patch[target_reference].updated=0;

#ifdef DEBUG
   fprintf(stderr,"Sending: %s",outbuffer);   
#endif

  return(0);
 }
 
 return(-1);
}
