/* 
    Threaded socket routines - main subroutines

    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

#include "netsocket.h"

/* Converts ascii text to in_addr struct */

struct in_addr *atoaddr(address)
char *address;
{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if (saddr.s_addr != -1) {
    return &saddr;
  }
    
  host = gethostbyname(address);
  if (host != NULL) {
    return (struct in_addr *) *host->h_addr_list;
  }
  return NULL;
}

/* Socket IO */
   
int sock_write (int sockfd, char *buf, size_t count)
{
 size_t bytes_sent = 0;
 int this_write;
 
 while (bytes_sent < count) {
  do {
   /* Write data */
   this_write = write(sockfd, buf, count - bytes_sent);
  } while ( (this_write < 0) && (errno == EINTR) );
  if (this_write <= 0) {
   return(this_write);
  } 
  bytes_sent += this_write;
  buf += this_write;
 }
 return(count);
}

int sock_read (int sockfd, char *buf, size_t count)
{
 size_t bytes_read = 0;
 int this_read;
 fd_set readers;
 struct timeval timeout;
 int pending;

 while (bytes_read < count) {
  do {
   /* Check status of connection */
   pending=0;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=5;
    timeout.tv_usec=0;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if (pending<0) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
   } 
   this_read = read(sockfd, buf, count - bytes_read);
  } while ( (this_read < 0) && (errno == EINTR) );
  if ((this_read==0) && (pending==1)) {
   fprintf (stderr,"Connection closed ...\n");
   return(-1);
  }    
  if (this_read <= 0) {
   return(this_read);
  } 
  bytes_read += this_read;
  buf += this_read;
 }
 return(count);
}
      
int sock_gets (int sockfd, char *str, size_t count)
{
 int bytes_read;
 int total_count = 0;
 char *current_position;
 char last_read = 0;
 fd_set readers;
 struct timeval timeout;
 int pending;
          
 current_position = str;
 while ((last_read != 10) && (last_read !=13)) {
   /* Check status of connection */
   pending=0;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=5;
    timeout.tv_usec=0;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if (pending<0) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
   } 
   bytes_read = read(sockfd, &last_read, 1);
   if ((bytes_read==0) && (pending==1)) {
    fprintf (stderr,"Connection closed ...\n");
    return(-1);
   }    
   if (bytes_read <= 0) {
    return bytes_read; 
   }
   if ( (total_count < count) && (last_read != 10) && (last_read !=13) ) {
    current_position[0] = last_read;
    current_position++;
    total_count++;
   }
 }
 if (count > 0) current_position[0] = 0;
 return total_count;
}

/* Network service thread                 */

#define BUFFER_LEN	128

void NetworkThread(struct netthread *nt)
{
 struct sockaddr_in address;
 int listening_socket = -1;
 int connected_socket = -1;
 int connected = 0;
 int parse_result;
 char inbuffer[BUFFER_LEN];
 char outbuffer[BUFFER_LEN];
 int length;
 char hostname[128];
 int retval;
 int optarg=1;
 
 /* Setup internet address information */  
 memset((char *) &address, 0, sizeof(address));
 gethostname(hostname,sizeof(hostname));
 address.sin_family = AF_INET;
 address.sin_port = htons(nt->port);
 address.sin_addr.s_addr = INADDR_ANY;

 /* Create listening socket */
 listening_socket = socket(AF_INET, SOCK_STREAM, 0);
 if (listening_socket < 0) {
  perror("socket");
  exit(-1);
 }

 /* Set socket options */
 if (setsockopt(listening_socket,SOL_SOCKET,SO_REUSEADDR, &optarg, sizeof(optarg))==-1) {
  perror("setsockopt");
  exit(-1);
 }

 /* Bind socket to address */
 if (bind(listening_socket,(struct  sockaddr*)&address,sizeof(address)) < 0) {
  perror("bind");
  close(listening_socket);
  exit(-1);
 }
 /* Allow only a single connection */
 listen(listening_socket, 1);
 
 /* Get a connection and handle it */
 while (1) {
  if (nt->verbose) {
   fprintf (stderr,"Waiting for connection on port %i.\n",nt->port);
  }
  length=sizeof(address);
  connected_socket = accept(listening_socket,(struct sockaddr *)&address, &length);
  if ((connected_socket < 0) && (errno != EINTR)) {
   perror("accept");
   exit(-1);
  } else {
   if (nt->verbose) {
    fprintf (stderr,"Connected ...\n");
   } 
   connected=1;
   while (connected) {
    retval=sock_gets(connected_socket,inbuffer,BUFFER_LEN);
    if (retval>0) {
     if (inbuffer[0]=='l') {
      if (nt->verbose) {
       fprintf (stderr,"Logout.\n");
      } 
      connected=0;
     } else {
      /* Lock chared data */
      shared_data_lock(nt);
      /* Set data using ___external___ routine */
      parse_result=command_to_output(inbuffer,outbuffer,BUFFER_LEN);
      /* Unlock shared data */
      shared_data_unlock(nt);
      /* Send result */
      if (parse_result!=-1) {
       /* Send data to network socket */
       retval=sock_write(connected_socket,outbuffer,strlen(outbuffer));
       connected=(retval>0);
      } 
     } 
    } else if (retval<0) {
     connected=0;
    }   
   }
   close(connected_socket);
  }
 }
 close(listening_socket);
 exit(0);
}

int shared_data_lock(struct netthread *nt)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&nt->network_mutex);
 
 return(0);
}

int shared_data_unlock(struct netthread *nt)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&nt->network_mutex);
 
 return(0);
}

int start_network_socket(struct netthread *nt, 
			unsigned short port,
			int verbose)
{
 /* Set port and other variables */
 nt->port=port;
 nt->verbose=verbose;
 
 /* Initialize sycro mutex */
 if (pthread_mutex_init(&nt->network_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }
       
 /* Start grabbing thread */
 if (pthread_create(&nt->network_thread, NULL,(void *)NetworkThread,(void *)nt)==-1) {
  perror ("pthread_create()");
  return(-1);
 } else {
  return(0);
 } 
}

int stop_network_socket(struct netthread *nt)
{
 /* Stop grabbing thread */
 if (pthread_cancel(nt->network_thread)==-1) {
  perror("pthread_cancel()");
  return(-1);
 } else {
  return(0);
 } 
}
