/*

    Program: Tracks motion in rectangular or irregular patches
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  | 
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/*
 *
 * Requires XWindows running at 16bpp color depth.
 * Requires libbgrab-2.1 or greater
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <limits.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <errno.h>

#include "bgrab.h"
#include "xutil.h"
#include "filters.h"

/* Framegrabber and screen color resolution */

#define DEPTH 	1
#define XDEPTH  2

/* String length limit for this program */

#define LINE_LEN	2048

/* Image saving counter */

int capturecount=0;

/* Largest number of areas that can be tested */

#define MAX_PATCH_NUM	1024

#define PATCH_NONE		0
#define PATCH_RECTANGULAR	1
#define PATCH_SHAPE		2

#define MAX_NAMELENGTH		32

/* Areas to test for difference changes */

struct tpatch {
		 /* From configuration */
		 char name[MAX_NAMELENGTH];    /* Patch identifier/filename */
		 int type;		       /* Type of patch - autodetect */
		 int xmin, xmax;	       /* X limits of patch - autodetect */
		 int ymin, ymax;	       /* Y limits of patch - autodetect */
		 unsigned char *shapeimage;    /* Shape image */
		 float ifheight,ifwidth;       /* inverse height and width */
                 /* Calculated */
                 int tracked;                  /* Is the location updated */
		 float posx, posy;	       /* Tracked absolute location */
		 float oldposx, oldposy;       /* Old tracked absolute location */
		 float direction;	       /* Angle of tracked motion */
		} patch[MAX_PATCH_NUM];

/* Patch count */

static int patchnum=0;

/* Color table for output graphics */

#define NUM_COLORTABLE	128

static int colortable[NUM_COLORTABLE];

/* Storage queue for "old" frames */

#define MAX_QUEUEDEPTH	256

static char *imagequeue, *curqueue;
static int curqueuenum;

/* Storage area for difference frames */

static char *differenceframe;
static int numdifferenceframe=0;
static int differenceframeindex[MAX_QUEUEDEPTH];

/* Summing frames */

static char *combinedframe;

/* Thresholded combinedframe */

static char *thresholdframe;

/* master masking frame */

static char *maskimage;

/* Shaped thresholdframe */

static char *shapeframe;

/* Command line configurable variables */

static int WIDTH, HEIGHT;
static int INPUTCHANNEL;
static int VIDEOMODE; 
static int REGION; 
static int CHANNEL;
static int QUEUEDEPTH;
static int THRESHOLD;
static int CONTRAST;
static int BRIGHTNESS;
static int DISPLAY_ON;
static int WHICH_DISPLAY;
static int TRIGGERLIMIT;
static float DIRLIMIT;
static int REFLECTORMODE;
static int FPSMODE;
static unsigned short PORT;
static char HOSTNAME[LINE_LEN];
static int NOX;
static int BINARIZE;

/* Videodevice to connect to */
static char VIDEODEV[LINE_LEN];

/* String that contains the frame numbers to difference */ 
static char QUEUEFRAMES[LINE_LEN];  

/* String that contains the filename for the master mask */
static char MASKNAME[LINE_LEN];

/* String that contains the filenames for the patches */
static char PATCHNAMES[LINE_LEN];

/* Path to patches for the patches */
static char PATCHPATH[LINE_LEN];

void set_defaults()
{
 WIDTH=640;
 HEIGHT=480;
 INPUTCHANNEL=CHANNEL_TUNER;
 strcpy (VIDEODEV,"/dev/video");
 VIDEOMODE=VIDEOMODE_NTSC;
 REGION=REGION_NTSC_CABLE;
 CHANNEL=1;
 QUEUEDEPTH=2;
 strcpy (QUEUEFRAMES,"1");
 DISPLAY_ON=1;
 WHICH_DISPLAY=0;
 THRESHOLD=128;
 CONTRAST=128;
 BRIGHTNESS=128;
 strcpy (MASKNAME,"");
 strcpy (PATCHNAMES,"sample");
 strcpy (PATCHPATH,"");
 TRIGGERLIMIT=2;
 DIRLIMIT=3.0*3.0;
 REFLECTORMODE=0;
 FPSMODE=1;
 PORT=0;
 strcpy(HOSTNAME,"");
 NOX=0;
 BINARIZE=0;
}

/* Load RGB colors from a file, fill table and randomize it consistently */

void loadColortable(char *filename)
{
 int i,r,g,b;
 FILE *inputfile;
 char aline[LINE_LEN];
 
 /* Initialize table with red */
 for (i=0; i<NUM_COLORTABLE; i++) {
  colortable[i]=0xf800;
 }
 /* Open file */
 if ((inputfile=fopen(filename,"r"))==NULL) {
  return;
 }
 /* Read unique colors from file */
 i=0;
 strcpy (aline,"");
 fgets(aline,LINE_LEN,inputfile);
 while ( (i<NUM_COLORTABLE) && (!(feof(inputfile))) ) {
  sscanf(aline,"%i %i %i",&r,&g,&b);
  colortable[i]=((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3); 
  fgets(aline,LINE_LEN,inputfile);
  i++;
 } 
 fclose(inputfile);
}

/* Drawing tables and routines */

static float COS_TABLE[360], SIN_TABLE[360];

void InitializeSinCosTable(void)
{
 int i;
 double angle;
 
 for(i=0;i<360;i++) {
  angle=(double)i/180.0*M_PI;
  COS_TABLE[i]=cos(angle);
  SIN_TABLE[i]=sin(angle);
 }
}

/* Load a .pgm (greyscale) image of screen size and return a pointer to it */

unsigned char *LoadShapeImage (char *filename)
{
 int xsize,ysize,numcolors,readitems;
 char pathname[LINE_LEN];
 char header[LINE_LEN];
 char errormessage[LINE_LEN];
 char aline[LINE_LEN];
 FILE *infile;
 unsigned char *tempbuffer;
  
 /* Build pathname */
 strcpy (pathname,PATCHPATH);
 if (strlen(PATCHPATH)>0) {
  strcat (pathname,"/");
 }
 strcat (pathname,filename);
 strcat (pathname,".pgm");
 
 /* Message */
 fprintf (stderr,"Loading patch image %s: ",pathname);

 /* Open and read file */
 if ((infile=fopen(pathname,"r"))==NULL) {
  sprintf(errormessage,"fopen(%s)",pathname);
  perror(errormessage);
  return(NULL);
 }

 /* Read header */
 fgets(aline,LINE_LEN,infile);
 sscanf(aline,"%s",header);

 /* Dump comment lines */
 fgets(aline,LINE_LEN,infile);
 while (aline[0]=='#') {
  fgets(aline,LINE_LEN,infile);
 } 

 /* Convert size from line */
 sscanf(aline,"%i %i",&xsize,&ysize);

 /* Convert depth from line */ 
 fgets(aline,LINE_LEN,infile);
 sscanf(aline,"%i",&numcolors);

 /* Check image size */
 if ((xsize==WIDTH) && (ysize==HEIGHT)) {
  /* Get memory */
  tempbuffer=malloc(WIDTH*HEIGHT*DEPTH);
  /* Read data */
  if ((readitems=fread(tempbuffer,(size_t)(WIDTH*HEIGHT*DEPTH),1,infile))==0) {
   free(tempbuffer);
   fprintf (stderr,"Error reading %s!\n",pathname);
   fclose(infile);
   return(NULL);
  } 
 } else {
  fprintf (stderr,"Error reading %s: Bad Size (%ix%i)!\n",filename,xsize,ysize);
  fclose(infile);
  return(NULL);
 } // size check 

 fclose(infile);
 fprintf (stderr," read %i bytes.\n",readitems*(WIDTH*HEIGHT*DEPTH));   
 
 return(tempbuffer); 
}

/* ---- Network routines ------- */

/* Converts ASCII text to in_addr struct */

struct in_addr *atoaddr(address)
char *address;
{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if (saddr.s_addr != -1) {
    return &saddr;
  }
    
  host = gethostbyname(address);
  if (host != NULL) {
    return (struct in_addr *) *host->h_addr_list;
  }
  return NULL;
}

/* Write buffer to a socket connection */
   
int sock_write (int sockfd, char *buf, size_t count)
{
 size_t bytes_sent = 0;
 int this_write;
       
 while (bytes_sent < count) {
  do {
   this_write = write(sockfd, buf, count - bytes_sent);
  } while ( (this_write < 0) && (errno == EINTR) );
  if (this_write <= 0) {
   return(this_write);
  } 
  bytes_sent += this_write;
  buf += this_write;
 }
 return(count);
}

/* -------------- Main routine --------------- */

int main (int argc, char *argv[])
{
 int i,j,x,y,framecount;
 struct fgdevice fg;
 struct xwinbuffer xwin;
 int not_first_time=0;
 int in_loop=1;
 KeySym key;
 int button=0;
 int mousex, mousey;
 int captureflag=0;
 char cfilename[LINE_LEN];
 FILE *capturefile;
 int curposnum;
 unsigned char *curposframe=NULL;
 unsigned short *curposscreen=NULL;
 const char delimiters[] = " .,:;";
 char *running,*token;
 int xmin,xmax,ymin,ymax,pcount;
 struct tpatch *curpatch;
 unsigned char *curposshape=NULL;
 unsigned char *curposscan=NULL;
 long sumn, summ, sumx, sumy;
 float dx,dy,isumn,ifwidth,ifheight;
 int linecolor;
 struct sockaddr_in address;
 static struct in_addr *addr;
 int connected;
 int connected_socket = -1;
 char outputline[LINE_LEN];
 char refstring[LINE_LEN];
 int display_sel,max_display_sel,display_selmode;
 unsigned char *displaybuffer=NULL;
    
 /* -------- Program Title */ 
 fprintf (stderr,"tracker - motion tracker, A. Schiffler, 04/00-05/00\n\n");
     
 /* -------- Parse command line */
 set_defaults();
 for (i=1; i<argc; i++) {
  char	*arg;
  if (argv[i][0]!='-') break;
  arg = argv[i];
  arg++;
  if (*arg == '-') arg++;
  if (strncmp(arg,"width",5)==0) {
   i++; 
   if (i<=argc) {
    WIDTH = atoi(argv[i]);
   }
  } else if (strncmp(arg,"height",6)==0) {
   i++; 
   if (i<=argc) {
    HEIGHT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"input",5)==0) {
   i++; 
   if (i<=argc) {
    INPUTCHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"region",6)==0) {
   i++; 
   if (i<=argc) {
    REGION = atoi(argv[i]);
   }
  } else if (strncmp(arg,"channel",7)==0) {
   i++; 
   if (i<=argc) {
    CHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"videodev",8)==0) {
   i++; 
   if (i<=argc) {
    strcpy (VIDEODEV,argv[i]);
   }
  } else if (strncmp(arg,"pal",3)==0) {
   VIDEOMODE=VIDEOMODE_NTSC;
  } else if (strncmp(arg,"ntsc",4)==0) {
   VIDEOMODE=VIDEOMODE_NTSC;
  } else if (strncmp(arg,"secam",5)==0) {
   VIDEOMODE=VIDEOMODE_SECAM;
  } else if (strncmp(arg,"binarize",8)==0) {
   BINARIZE=1;
  } else if (strncmp(arg,"display",7)==0) {
   DISPLAY_ON=1;
   i++; 
   if (i<=argc) {
    WHICH_DISPLAY = atoi(argv[i]);
   } else {
    WHICH_DISPLAY = 0;
   }
  } else if (strncmp(arg,"nodisplay",9)==0) {
   DISPLAY_ON=0;
  } else if (strncmp(arg,"queuedepth",10)==0) {
   i++; 
   if (i<=argc) {
    QUEUEDEPTH = atoi(argv[i]);
    if (QUEUEDEPTH>MAX_QUEUEDEPTH) {
     QUEUEDEPTH=MAX_QUEUEDEPTH;
    }
   }
  } else if (strncmp(arg,"threshold",9)==0) {
   i++; 
   if (i<=argc) {
    THRESHOLD = atoi(argv[i]);
    if (THRESHOLD>255) {
     THRESHOLD=255;
    } else if (THRESHOLD<0) {
     THRESHOLD=0;
    }
   }
  } else if (strncmp(arg,"contrast",8)==0) {
   i++; 
   if (i<=argc) {
    CONTRAST = atoi(argv[i]);
    if (CONTRAST>255) {
     CONTRAST=255;
    } else if (CONTRAST<0) {
     CONTRAST=0;
    }
   }
  } else if (strncmp(arg,"brightness",10)==0) {
   i++; 
   if (i<=argc) {
    BRIGHTNESS = atoi(argv[i]);
    if (BRIGHTNESS>255) {
     BRIGHTNESS=255;
    } else if (BRIGHTNESS<0) {
     BRIGHTNESS=0;
    }
   }
  } else if (strncmp(arg,"triggerlimit",12)==0) {
   i++; 
   if (i<=argc) {
    TRIGGERLIMIT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"dirlimit",8)==0) {
   i++; 
   if (i<=argc) {
    DIRLIMIT = atof(argv[i]);
   }
  } else if (strncmp(arg,"queueframes",11)==0) {
   i++; 
   if (i<=argc) {
    strncpy (QUEUEFRAMES,argv[i],LINE_LEN);
   }
  } else if (strncmp(arg,"patchpath",9)==0) {
   i++; 
   if (i<=argc) {
    strncpy (PATCHPATH,argv[i],LINE_LEN);
   }
  } else if (strncmp(arg,"patchnames",10)==0) {
   i++; 
   if (i<=argc) {
    strncpy (PATCHNAMES,argv[i],LINE_LEN);
   }
  } else if (strncmp(arg,"mastermask",11)==0) {
   i++; 
   if (i<=argc) {
    strncpy (MASKNAME,argv[i],LINE_LEN);
   }
  } else if (strncmp(arg,"reflector",9)==0) {
   REFLECTORMODE=1;
  } else if (strncmp(arg,"nofps",5)==0) {
   FPSMODE=0;
  } else if (strncmp(arg,"nox",3)==0) {
   NOX=1;
  } else if (strncmp(arg,"port",4)==0) {
   i++; 
   if (i<=argc) {
    PORT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"hostname",8)==0) {
   i++; 
   if (i<=argc) {
    strcpy (HOSTNAME,argv[i]);
   }
  } else {
   /* Command line help */
   printf ("Command line parameters:\n");
   printf ("\nVideo settings:\n");
   printf (" -width #           width of images (default: %i)\n",WIDTH);
   printf (" -height #          height of images (default: %i)\n",HEIGHT);
   printf (" -input #           input channel (default: %i)\n",INPUTCHANNEL);
   printf (" -videodev txt      video device (default: %s)\n",VIDEODEV);
   printf (" -pal               use PAL video (default: ntsc)\n");
   printf (" -ntsc              use NTSC video (default: ntsc)\n");
   printf (" -secam             use SECAM video (default: ntsc)\n");
   printf (" -region #          tuner frequency region (default: %i)\n",REGION);
   printf (" -channel           tuner channel (default: %i)\n",CHANNEL);
   printf (" -contrast #        contrast setting (0-255) (default %i)\n",CONTRAST);
   printf (" -brightness #      brightness setting (0-255) (default %i)\n",BRIGHTNESS);
   printf ("\nOutput settings:\n");
   printf (" -display #         live display set to mode # (default: 0)\n");
   printf (" -nodisplay         no display of live video or patches\n");
   printf (" -reflector         turn on asciireflector PUSH output mode (default: off)\n");
   printf (" -nofps             turn off FPS reporting\n");
   printf (" -nox               don't even open an XWindow\n");
   printf ("\nDetection settings:\n");
   printf (" -queuedepth #      length of imagequeue (>0) (default: %i)\n",QUEUEDEPTH);
   printf (" -queueframes txt   list of frame#s to difference the current\n");
   printf ("                    frame 0 against (default: %s)\n",QUEUEFRAMES);
   printf (" -binarize          do binarization on difference (default: off)\n");
   printf (" -threshold #       binarization limit/threshold (0-255) (default: %i)\n",THRESHOLD);
   printf (" -triggerlimit #    number of pixels to trigger patch (default: %i)\n",TRIGGERLIMIT);
   printf (" -dirlimit #.#      minimum distance between positions before new\n");
   printf ("                    direction is calculated (default: %4.1f)\n",DIRLIMIT);
   printf ("\nPatch settings:\n");
   printf (" -patchpath txt     path to patchfiles (default: local directory)\n");
   printf (" -patchnames txt    filenames of .pgm images (default: %s)\n",PATCHNAMES);   
   printf (" -mastermask txt    filename of .pgm image that servers as a master mask\n");
   printf ("                    tp be applied before tracking (default: none)\n");
   printf ("\nNetwork Settings:\n");
   printf (" -hostname txt      hostname to connect to (default: don't connect)\n");
   printf (" -port #            port to connect to (default: %i)\n",PORT);   
   exit(0);
  }
 }
 
 /* Sin/Cos table for drawing routines */
 InitializeSinCosTable();

 /* Set and load colortable */
 loadColortable("rgb.txt");
 
 /* Current and maximum display selection, selection mode */
 display_sel=0;
 display_selmode=0;
  
 /* Inverse sizes for conversions */
 ifwidth=1.0/(float)WIDTH;
 ifheight=1.0/(float)HEIGHT;

 /* Internet connection */
 if ((PORT!=0) && (strlen(HOSTNAME)>0)) {
  /* Setup internet address information */
  addr = (struct in_addr *)atoaddr((char *)HOSTNAME);
  if (addr == NULL) {
   fprintf(stderr,"Invalid network address: %s",HOSTNAME);
   exit(-1);
  }
  memset((char *) &address, 0, sizeof(address));
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = addr->s_addr;
  address.sin_port = htons(PORT);

  /* Connect to server */ 
  connected_socket = socket(AF_INET, SOCK_STREAM, 0);
  printf("Connecting to %s on port %d ...",inet_ntoa(*addr),PORT);
  connected = connect(connected_socket, (struct sockaddr *) &address, sizeof(address));
  if (connected < 0) {
   printf("\n");
   perror("connect()");
   exit(-1);
  }
  printf(" connected.\n");
  connected=1;
 } else {
  connected=0;
 }
  
 /* ---------- Checks and messages */
 fprintf (stderr, "Using Device: %s, Channel: %i, Size: %ix%i.\n",VIDEODEV, INPUTCHANNEL, WIDTH, HEIGHT);
 
 /* Setup and check queueframes */
 numdifferenceframe=0;
 running=QUEUEFRAMES;
 while ((token=strsep(&running,delimiters))!=NULL) {
  differenceframeindex[numdifferenceframe]=atoi(token);
  if ((differenceframeindex[numdifferenceframe]<QUEUEDEPTH) && (differenceframeindex[numdifferenceframe]>0)) {
   numdifferenceframe++;
  }
 } 
 if (numdifferenceframe==0) {
  fprintf (stderr,"Error: At least one queueframe for differencing required.\n");
  exit(-1);
 }
 fprintf (stderr, "Differencing queueframes:");
 for (i=0; i<numdifferenceframe; i++) {
  if (i!=0) printf (",");
  fprintf (stderr, " %i",differenceframeindex[i]); 
 }
 fprintf (stderr, "\n");

 /* -------  Load and process shape images */
 patchnum=0;
 running=PATCHNAMES;
 while ((token=strsep(&running,delimiters))!=NULL) {
  if ((patch[patchnum].shapeimage=LoadShapeImage(token)) != NULL) {    
   /* Call MMX assembly routine to binarize the image (in-place) */
   binarize(patch[patchnum].shapeimage,patch[patchnum].shapeimage,(WIDTH*HEIGHT*DEPTH),(unsigned char)254);
   /* Determine bounds and calculate number of pixels in shape */
   curposshape=patch[patchnum].shapeimage;
   xmin=WIDTH; xmax=-1;
   ymin=HEIGHT; ymax=-1;
   pcount=0;
   for (y=0; y<HEIGHT; y++) {
    for (x=0; x<WIDTH; x++) {
     if (*curposshape==255) {
      if (x<xmin) xmin=x; 
      if (x>xmax) xmax=x;
      if (y<ymin) ymin=y; 
      if (y>ymax) ymax=y;
      pcount++;
     }
     curposshape++;
    }
   }
   /* Store name, bounds and determine type */
   strcpy(patch[patchnum].name,token);
   patch[patchnum].xmin=xmin;
   patch[patchnum].xmax=xmax;
   patch[patchnum].ymin=ymin;
   patch[patchnum].ymax=ymax;
   patch[patchnum].ifwidth=1.0/(float)(xmax-xmin+1);
   patch[patchnum].ifheight=1.0/(float)(ymax-ymin+1);
   if (((xmax-xmin+1)*(ymax-ymin+1))==pcount) {
    patch[patchnum].type=PATCH_RECTANGULAR;
    /* Delete the image again */
    free(patch[patchnum].shapeimage);
    patch[patchnum].shapeimage=NULL;    
   } else {
    patch[patchnum].type=PATCH_SHAPE;
   }
   /* Report */
   fprintf (stderr,"Added patch #%i: name '%s' bounds %ix%i-%ix%i type %i\n",patchnum+1,token,xmin,ymin,xmax,ymax,patch[patchnum].type);
   /* Next patch */     
   patchnum++;
  } 
 } 
 if (patchnum==0) {
  fprintf (stderr,"Error: At least one patch is required.\n");
  exit(-1);
 }

 /* Load master mask if required */
 maskimage=NULL;
 if (strlen(MASKNAME)>0) {
  if ((maskimage=LoadShapeImage(MASKNAME)) != NULL) {    
   /* Call MMX assembly routine to binarize the image (in-place) */
   binarize(maskimage,maskimage,(WIDTH*HEIGHT*DEPTH),(unsigned char)254);
  }
 }
      
 /* Framegrabber initialization */
 if (fg_open_device (&fg, VIDEODEV)!=0) {
  exit(-1); 
 }
 
 /* Grabber info */	
 if (fg_print_info (&fg)!=0) {
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 }
 
 /* Channel setup */
 if (fg_set_channel (&fg, INPUTCHANNEL, VIDEOMODE)!=0) {
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 } 
 
 /* Tuner setup */
 if (fg_set_frequency (&fg, REGION, CHANNEL)!=0) {
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 } 

 /* Create storage queue */ 
 if ((imagequeue=(char *)malloc(QUEUEDEPTH*(WIDTH*HEIGHT*DEPTH)))==NULL) {
  fprintf(stderr,"Not enough memory for a %i frame queue (%i bytes) - exiting.\n",QUEUEDEPTH,QUEUEDEPTH*(WIDTH*HEIGHT*DEPTH));
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 } else {
  curqueue=imagequeue;
  curqueuenum=0;
 }

 /* Create difference storage */
 if ((differenceframe=(unsigned char *)malloc(WIDTH*HEIGHT*DEPTH))==NULL) {
  fprintf(stderr,"Not enough memory for differenceframe image (%i bytes) - exiting.\n",(WIDTH*HEIGHT*DEPTH));
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 }

 /* Create combined storage */
 if ((combinedframe=(unsigned char *)malloc(WIDTH*HEIGHT*DEPTH))==NULL) {
  fprintf(stderr,"Not enough memory for combinedframe image (%i bytes) - exiting.\n",(WIDTH*HEIGHT*DEPTH));
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 }

 /* Create threshold storage */
 if ((thresholdframe=(unsigned char *)malloc(WIDTH*HEIGHT*DEPTH))==NULL) {
  fprintf(stderr,"Not enough memory for thresholdframe image (%i bytes) - exiting.\n",(WIDTH*HEIGHT*DEPTH));
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 }

 /* Create shape-stencil storage */
 if ((shapeframe=(unsigned char *)malloc(WIDTH*HEIGHT*DEPTH))==NULL) {
  fprintf(stderr,"Not enough memory for shapeframe image (%i bytes) - exiting.\n",(WIDTH*HEIGHT*DEPTH));
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 }

 /* Create displaybuffer storage */
 if ((displaybuffer=(unsigned char *)malloc(WIDTH*HEIGHT*DEPTH))==NULL) {
  fprintf(stderr,"Not enough memory for displaybuffer image (%i bytes) - exiting.\n",(WIDTH*HEIGHT*DEPTH));
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 }

 /* Set settings */
 fg_set_setting(&fg,SETTING_CONTRAST,CONTRAST*256);
 fg_set_setting(&fg,SETTING_BRIGHTNESS,BRIGHTNESS*256);

 /* Do we need Xwindows */  
 if (NOX==0) {
  /* Setup xwindows */
  if (setup_xwinbuffer (&xwin, argv[0], WIDTH, HEIGHT, XDEPTH)!=0) {
   if (fg_close_device (&fg)!=0) exit(-1); 	
   exit(-1);
  }
  /* Write something if interactive screen output is turned off */
  if (DISPLAY_ON==0) {
   /* Blank screen and draw message */
   memset(xwin.shmimage->data,0,(WIDTH*HEIGHT*XDEPTH));
   drawstring (&xwin, 8, 8, "Display OFF", 0xffffff);     
   /* Draw screen */
   sync_xwinbuffer(&xwin);
  }
 }
 
 /* Setup continous grabbing */
 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_GREY)!=0) {
  if (close_xwinbuffer(&xwin)!=0) exit(-1);
  if (fg_close_device (&fg)!=0) exit(-1); 	
  exit(-1);
 } 
 
 /* FPS reporting interval */
 fg_set_fps_interval(&fg,100);
 
 /* Setup reflector string */
 if (REFLECTORMODE) {
  /* Output a PUSH for the reflector-mode */
  strcpy(refstring,"PUSH");
 } else {
  strcpy(refstring,"");
 }

 /* Current and maximum display selection */
 display_sel=0;
 if (patchnum>numdifferenceframe) {
  max_display_sel=patchnum;
 } else {
  max_display_sel=numdifferenceframe;
 }
 
 /* View/processing loop */
 framecount=0;
 while (in_loop) { 
  /* Keep counting frames */
  framecount++;
  
  /* Update framebuffer with new image */
  if (fg_get_next_image(&fg)!=NULL) {

    /* ---- Image capture ---- */
    
    /* Image capture to file on user request */
    if (captureflag) {
     sprintf(cfilename,"capture%i.pgm",capturecount);
     printf ("Writing %s ...\n",cfilename);
     capturefile=fopen(cfilename,"w");
     /* Write headers */
     fprintf(capturefile,"P5\n# libbgrab tracker\n%i %i\n255\n",WIDTH,HEIGHT);
     /* Dump screen to file */
     fwrite ((void *)fg.current_image,WIDTH*HEIGHT,1,capturefile);
     /* Close file and update variables */
     fclose(capturefile);
     capturecount++;
     captureflag=0;
    } 

    /* ---- Image storage queue ---- */
    
    /* Update queue pointer */
    if (curqueuenum==0) {
     curqueuenum=QUEUEDEPTH-1;
     curqueue = imagequeue;
     curqueue += (fg.image_size*(QUEUEDEPTH-1));
    } else {
     curqueuenum--;
     curqueue -= (fg.image_size);
    }
   
    /* Copy image to queue */
    memcpy((char *)curqueue,(char *)fg.current_image,fg.image_size);

    /* ---- Difference processing ----- */
        
    /* Clear adding-storage area */
    bzero(combinedframe,(HEIGHT*WIDTH*DEPTH));
    /* Use assembly difference routine to calculate absolute difference */
    for (i=0; i<numdifferenceframe; i++) {
     /* Pointer setup */
     curposnum=((curqueuenum + differenceframeindex[i]) % QUEUEDEPTH);
     curposframe=imagequeue;
     curposframe += (WIDTH*HEIGHT*DEPTH*curposnum);
     /* Call MMX assembly routine to calculate absolute difference */
     absdiff(fg.current_image,curposframe,differenceframe,(WIDTH*HEIGHT*DEPTH));
     /* Maybe copy this differenceframe for display */
     if ((DISPLAY_ON) && (WHICH_DISPLAY==1) && (i==(display_sel % numdifferenceframe))) {
      memcpy(displaybuffer,differenceframe,(WIDTH*HEIGHT*DEPTH));
     }
     /* Call MMX assembly routine to add the differences */
     addbytes(differenceframe,combinedframe,combinedframe,(WIDTH*HEIGHT*DEPTH));
    }
    if (BINARIZE) {
     /* Call MMX assembly routine to threshold the sum and binarize the image */
     binarize(combinedframe,thresholdframe,(WIDTH*HEIGHT*DEPTH),(unsigned char)THRESHOLD);
    } else {
     /* Call MMX assembly routine to threshold the sum and binarize the image */
     threshold(combinedframe,thresholdframe,(WIDTH*HEIGHT*DEPTH),(unsigned char)THRESHOLD);
    }
    /* Apply master masking routine */
    if (maskimage != NULL) {
     /* Call MMX assembly routine to stencil the image with the shape */
     andbytes(thresholdframe,maskimage,thresholdframe,(WIDTH*HEIGHT*DEPTH));
    }
    
    /* ---- Patch processing ----- */
    
    curpatch=patch; 
    for (i=0; i<patchnum; i++) {
     if (curpatch->type==PATCH_RECTANGULAR) {
      /* Scan thresholdframe */
      curposscan=thresholdframe;
      /* Maybe blank displaybuffer */
      if ((DISPLAY_ON) && (WHICH_DISPLAY==5) && (i==(display_sel % patchnum))) {
       memset(displaybuffer,0,(WIDTH*HEIGHT*DEPTH));
      }
     } else {
      /* Call MMX assembly routine to stencil the image with the shape */
      andbytes(thresholdframe,curpatch->shapeimage,shapeframe,(WIDTH*HEIGHT*DEPTH));
      /* Maybe copy this shapeframe for display */
      if ((DISPLAY_ON) && (WHICH_DISPLAY==5) && (i==(display_sel % patchnum))) {
       memcpy(displaybuffer,shapeframe,(WIDTH*HEIGHT*DEPTH));
      }
      curposscan=shapeframe;
     }

     /* Scan and sum according to binarization setting */
     if (BINARIZE) {
      /* Create tracking sums for binarized image */
      sumn=0;
      sumx=0;
      sumy=0;
      curposscan += (curpatch->xmin+curpatch->ymin*WIDTH);
      for (y=curpatch->ymin; y<=curpatch->ymax; y++) {
       for (x=curpatch->xmin; x<=curpatch->xmax; x++) {
        if (*curposscan!=0) {
         sumn++;
         sumx += (long)x;
         sumy += (long)y;
        }
        curposscan++;
       }
       curposscan += (WIDTH-curpatch->xmax+curpatch->xmin-1);
      }
      /* Calculate average and update other variables */
      if ((sumn>0) && (sumn>TRIGGERLIMIT)) {
       curpatch->tracked=1;
       curpatch->oldposx=curpatch->posx;
       curpatch->oldposy=curpatch->posy;
       isumn=1.0/(float)sumn;
       curpatch->posx=(float)sumx*isumn;
       curpatch->posy=(float)sumy*isumn;
       dx=curpatch->oldposx-curpatch->posx;
       dy=curpatch->oldposy-curpatch->posy;
       if ((dx*dx+dy*dy)>DIRLIMIT) {
        curpatch->direction=atan2((double)dy,(double)dx)*(float)(180.0/M_PI)+180.0;
       }
      } else {
       /* Not enough pixels for tracking */
       curpatch->tracked=0;
      } 
     } else {
      /* Create tracking sums for greyscale thresholded image */
      sumn=0;
      summ=0;
      sumx=0;
      sumy=0;
      curposscan += (curpatch->xmin+curpatch->ymin*WIDTH);
      for (y=curpatch->ymin; y<=curpatch->ymax; y++) {
       for (x=curpatch->xmin; x<=curpatch->xmax; x++) {
        if (*curposscan!=0) {
         sumn += (long)*curposscan;
         summ++;
         sumx += (long)*curposscan*(long)x;
         sumy += (long)*curposscan*(long)y;
        }
        curposscan++;
       }
       curposscan += (WIDTH-curpatch->xmax+curpatch->xmin-1);
      }
      /* Calculate average and update other variables */
      if ((sumn>0) && (summ>TRIGGERLIMIT)) {
       curpatch->tracked=1;
       curpatch->oldposx=curpatch->posx;
       curpatch->oldposy=curpatch->posy;
       isumn=1.0/(float)sumn;
       curpatch->posx=(float)sumx*isumn;
       curpatch->posy=(float)sumy*isumn;
       dx=curpatch->oldposx-curpatch->posx;
       dy=curpatch->oldposy-curpatch->posy;
       if ((dx*dx+dy*dy)>DIRLIMIT) {
        curpatch->direction=atan2((double)dy,(double)dx)*(float)(180.0/M_PI)+180.0;
       }
      } else {
       /* Not enough pixels for tracking */
       curpatch->tracked=0;
      } 
     }
     if (curpatch->tracked) {
      /* Generate output string */
      sprintf (outputline, "%s%s: AX=%5.3f AY=%5.3f RX=%5.3f RY=%5.3f DIR=%5.3f\n",
                  refstring,
                  curpatch->name,
                  curpatch->posx*ifwidth,
                  curpatch->posy*ifheight,
                  (float)((int)curpatch->posx-curpatch->xmin)*curpatch->ifwidth,
                  (float)((int)curpatch->posy-curpatch->ymin)*curpatch->ifheight,
                  curpatch->direction*(1.0/360.0));
      /* Output to screen */
      printf ("%s",outputline);
      /* Output to network */
      if (connected) {
       connected=(sock_write(connected_socket,outputline,strlen(outputline))>0);
      }
     }           
     /* Next patch */
     curpatch++;
    }
    /* Make sure everything gets written out */
    fflush(stdout);            
                    
    /* ---- Image/Results display ----- */
    
    if (NOX==0) {

     /* Draw screen */
     if (DISPLAY_ON) {       
      /* Copy an image to screenbuffer with 8bit-->16bit color conversion */
      /* Switch source depending on display mode */
      if (WHICH_DISPLAY==1) {
       /* The selected difference image */
       curposframe=(unsigned char *)displaybuffer;
      } else if (WHICH_DISPLAY==2) {
       /* The summed image */
       curposframe=(unsigned char *)combinedframe;
      } else if (WHICH_DISPLAY==3) {
       /* The thresholded image */
       curposframe=(unsigned char *)thresholdframe;
      } else if (WHICH_DISPLAY==4) {
       /* The valid shape image or live */
       if (patch[display_sel % patchnum].type==PATCH_SHAPE) {
        curposframe=(unsigned char *)patch[display_sel % patchnum].shapeimage;
       } else {
        /* Live image */
        curposframe=(unsigned char *)fg.current_image;
       } 
      } else if (WHICH_DISPLAY==5) {
       /* The last shaped threshold image */
       curposframe=(unsigned char *)displaybuffer;
      } else if (WHICH_DISPLAY==6) {
       if (maskimage!=NULL) {
        /* Master mask  */
        curposframe=(unsigned char *)maskimage;
       } else {
        /* Live image */
        curposframe=(unsigned char *)fg.current_image;
       } 
      } else {
       /* Live image */
       curposframe=(unsigned char *)fg.current_image;
      }
      if (WHICH_DISPLAY<6) {
       /* Copy y8 image to rgb565 screen */
       curposscreen=(unsigned short *)xwin.shmimage->data;
       for (i=0; i<fg.image_size; i++) {
        *curposscreen = fg.y8_to_rgb565[(int)*curposframe];
        curposscreen++;
        curposframe++;
       } 
      }
      /* Overlay display string */
      switch (WHICH_DISPLAY) {
       case 1:
        drawstring(&xwin, 8,8,"Difference",0x0000ffff);
        break;
       case 2:
        drawstring(&xwin, 8,8,"Summed Difference",0x0000ffff);
        break;
       case 3:
        drawstring(&xwin, 8,8,"Threshold Difference",0x0000ffff);
        break;
       case 4:
        drawstring(&xwin, 8,8,"Shape (or Live)",0x0000ffff);
        break;
       case 5:
        drawstring(&xwin, 8,8,"Shape Threshold (or Blank)",0x0000ffff);
        break;
       case 6:
        drawstring(&xwin, 8,8,"Master mask (or Live)",0x0000ffff);
        break;
       default: 
        drawstring(&xwin, 8,8,"Live",0x0000ffff);
        break;
      }
      /* Show selection number, patchname, type */
      if (patch[display_sel % patchnum].type==PATCH_SHAPE) {      
       sprintf(outputline,"Selection: #%i - %s (shape)",display_sel,patch[display_sel % patchnum].name);
      } else {
       sprintf(outputline,"Selection: #%i - %s (rectangle)",display_sel,patch[display_sel % patchnum].name);
      }
      drawstring(&xwin, 8,16, outputline,0x0000ffff);                 
      /* Overlay tracking results */
      curpatch=patch; 
      for (i=0; i<patchnum; i++) {
       /* Only draw one box in selection mode */
       if ( ( (display_selmode) && (i==(display_sel % patchnum)) ) || 
            ( (display_selmode==0) ) ) {
        /* Set color */
        linecolor=colortable[i % NUM_COLORTABLE];
        /* Patch box */
        hline(&xwin,curpatch->xmin,curpatch->xmin+3,curpatch->ymin,linecolor);
        hline(&xwin,curpatch->xmin,curpatch->xmin+3,curpatch->ymax,linecolor);
        hline(&xwin,curpatch->xmax-3,curpatch->xmax,curpatch->ymin,linecolor);
        hline(&xwin,curpatch->xmax-3,curpatch->xmax,curpatch->ymax,linecolor);
        vline(&xwin,curpatch->xmin,curpatch->ymin,curpatch->ymin+3,linecolor);
        vline(&xwin,curpatch->xmax,curpatch->ymin,curpatch->ymin+3,linecolor);
        vline(&xwin,curpatch->xmin,curpatch->ymax-3,curpatch->ymax,linecolor);
        vline(&xwin,curpatch->xmax,curpatch->ymax-3,curpatch->ymax,linecolor);
        /* Patch name */
        drawstring(&xwin,curpatch->xmin+3,curpatch->ymin+3,curpatch->name,linecolor);
        /* Set color to red/green depending on tracking results */
        if (curpatch->tracked) {
         linecolor=0x07d0;
        } else {
         linecolor=0xf800;
        }
        /* Tracking position */
        x=(int)curpatch->posx;
        y=(int)curpatch->posy;
        hline(&xwin,x-4,x+4,y,linecolor);
        vline(&xwin,x,y-4,y+4,linecolor);
        /* Show direction */
        for (j=0; j<360; j += 36) {
         drawpixel(&xwin,x+(int)(12.0*COS_TABLE[j % 360]), y+(int)(12.0*SIN_TABLE[j % 360]),linecolor);
        } 
        x += (int)(12.0*COS_TABLE[(int)curpatch->direction % 360]);
        y += (int)(12.0*SIN_TABLE[(int)curpatch->direction % 360]);
        hline(&xwin,x-1,x+1,y-1,linecolor);
        hline(&xwin,x-1,x+1,y,linecolor);
        hline(&xwin,x-1,x+1,y+1,linecolor);
       }      
       curpatch++;
      }
      /* Sync XWindow */
      sync_xwinbuffer(&xwin);
     }
    } 
  }

  if (NOX==0) {
  
   /* Poll keyboard - process mouse right after grabbing */
   if (keypoll_xwinbuffer (&xwin, &key, &button, &mousex, &mousey)) {

    switch (key) {
   
    case XK_q:
     in_loop=0;
     fprintf (stderr,"Exiting program\n");
     break; 

    case XK_s:
     captureflag=1;
     fprintf (stderr,"Capturing image\n");
     break; 

    case XK_F1:
     WHICH_DISPLAY=0;
     fprintf (stderr,"Showing normal video\n");
     break; 

    case XK_F2:
     WHICH_DISPLAY=1;
     fprintf (stderr,"Showing (selected) difference frame\n");
     break; 

    case XK_F3:
     WHICH_DISPLAY=2;
     fprintf (stderr,"Showing combined difference frames\n");
     break; 

    case XK_F4:
     WHICH_DISPLAY=3;
     fprintf (stderr,"Showing binarized frame\n");
     break; 

    case XK_F5:
     WHICH_DISPLAY=4;
     fprintf (stderr,"Showing (selected)  shape image\n");
     break; 

    case XK_F6:
     WHICH_DISPLAY=5;
     fprintf (stderr,"Showing (selected) shaped binarized image\n");
     break; 

   case XK_F7:
     WHICH_DISPLAY=6;
     fprintf (stderr,"Showing master mask image\n");
     break; 

    case XK_space:
     display_selmode = ((display_selmode + 1) % 2);
     fprintf (stderr,"Display selectionmode set to %i\n",display_selmode);
     break; 

    case XK_KP_Add:
     display_sel++;
     if (display_sel==max_display_sel) {
      display_sel=0;
     }
     fprintf (stderr,"Display selection is now %i\n",display_sel);
     break; 

    case XK_KP_Subtract:
     display_sel--;
     if (display_sel==-1) {
      display_sel=(max_display_sel-1);
     }
     fprintf (stderr,"Display selection is now %i\n",display_sel);
     break; 

    case XK_r:
     if (THRESHOLD>0) {
      THRESHOLD--;
      fprintf (stderr,"Threshold set to %i\n",THRESHOLD);
     }
     break; 
   
    case XK_t:
     if (THRESHOLD<255) {
      THRESHOLD++;
      fprintf (stderr,"Threshold set to %i\n",THRESHOLD);
     }
     break; 

    case XK_x:
     if (CONTRAST>0) {
      CONTRAST--;
      fprintf (stderr,"Contrast set to %i\n",CONTRAST);
      fg_set_setting(&fg,SETTING_CONTRAST,CONTRAST*256);
     }
     break; 
   
    case XK_c:
     if (CONTRAST<255) {
      CONTRAST++;
      fprintf (stderr,"Contrast set to %i\n",CONTRAST);
      fg_set_setting(&fg,SETTING_CONTRAST,CONTRAST*256);
     }
     break; 
   
    case XK_v:
     if (BRIGHTNESS>0) {
      BRIGHTNESS--;
      fprintf (stderr,"Brightness set to %i\n",BRIGHTNESS);
      fg_set_setting(&fg,SETTING_BRIGHTNESS,BRIGHTNESS*256);
     }
     break; 
   
    case XK_b:
     if (BRIGHTNESS<255) {
      BRIGHTNESS++;
      fprintf (stderr,"Brightness set to %i\n",BRIGHTNESS);
      fg_set_setting(&fg,SETTING_BRIGHTNESS,BRIGHTNESS*256);
     }
     break; 

    case XK_d:
     DISPLAY_ON++;
     DISPLAY_ON &= 1;
     if (DISPLAY_ON==0) {
      /* Blank screen and draw message */
      memset(xwin.shmimage->data,0,(WIDTH*HEIGHT*XDEPTH));
      drawstring (&xwin, 8, 8, "Display OFF", 0xffffff);     
      /* Draw screen */
      sync_xwinbuffer(&xwin);
     }
     fprintf (stderr,"Display_on set to %i\n",DISPLAY_ON);
     break; 

    case XK_f:
     FPSMODE++;
     FPSMODE &= 1;
     fprintf (stderr,"FPSmode set to %i\n",FPSMODE);
     break; 

    }
   }
  }
  
  /* Print position information on mouse-press */
  if (button) {
   fprintf (stderr, "Mouse is at X=%i Y=%i\n",mousex,mousey);
   button=0;
  } 
  
  /* Print fps */
  if (FPSMODE) {
   if (fg.framecount==0) {
    if (not_first_time) {
     fprintf (stderr, "FPS=%7.1f\n",fg_get_fps(&fg));
    } else {
     not_first_time=1;
    } 
   }
  }

 } // main loop

 if (NOX==0) {
  /* Close window */
  if (close_xwinbuffer(&xwin)!=0) exit(-1);
 }
 
 /* Stop framegrabber */
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	
 if (fg_close_device (&fg)!=0) exit(-1); 	

 /* Free memory */
 for (i=0; i<patchnum; i++) free(patch[i].shapeimage);
 free (combinedframe);
 free (differenceframe);
 free (thresholdframe);
 free (shapeframe);
 free (maskimage);
 
 /* Close network connection */
 if (connected) {
  close(connected_socket);
 }   
 
 exit(0);
}

