/*

    Program: webcam server
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

#include <stdio.h>
#include <stdlib.h>

#include "bgrab.h"
#include "webcamsocket.h"

/* Fixed at 16 bpp */

#define DEPTH 	2

/* Globals and default settings */

static int WIDTH, HEIGHT, CONNECTIONS, INPUTCHANNEL,
           PORT, JPEGGRAY, JPEGQUALITY, VIDEOMODE;
static char VIDEODEV[LINE_LEN];

void set_defaults()
{
 WIDTH=512;
 HEIGHT=384;
 CONNECTIONS=16;
 INPUTCHANNEL=CHANNEL_COMPOSITE;
 PORT=1331;
 strcpy (VIDEODEV,"/dev/video");
 JPEGGRAY=0;
 JPEGQUALITY=70;
 VIDEOMODE=VIDEOMODE_PAL;
}

/* Main program  */

int main (int argc, char *argv[])
{
 struct fgdevice fg;
 struct netthreads nt;
 unsigned short socket_port;
 struct sockaddr_in address;
 int listening_socket = -1;
 int connected_socket = -1;
 int in_loop;
 char hostname[128];
 int value;
 int length;
 int i;

 /* Message */
 printf ("webcam - bttv webcam server, (C) GPL, A. Schiffler, 12/1999\n\n");

 /* Initialize variables */
 set_defaults();
 
 /* Parse command line */
 for (i=1; i<argc; i++) {
  char	*arg;
  if (argv[i][0]!='-') break;
  arg = argv[i];
  arg++;
  if (*arg == '-') arg++;
  if (strcmp(arg,"width")==0) {
   i++; 
   if (i<=argc) {
    WIDTH = atoi(argv[i]);
   }
  } else if (strcmp(arg,"height")==0) {
   i++; 
   if (i<=argc) {
    HEIGHT = atoi(argv[i]);
   }
  } else if (strcmp(arg,"port")==0) {
   i++; 
   if (i<=argc) {
    PORT = atoi(argv[i]);
   }
  } else if (strcmp(arg,"maxcon")==0) {
   i++; 
   if (i<=argc) {
    CONNECTIONS = atoi(argv[i]);
   }
  } else if (strcmp(arg,"gray")==0) {
   i++; 
   if (i<=argc) {
    JPEGGRAY = atoi(argv[i]);
    JPEGGRAY &= 1;
   }
  } else if (strcmp(arg,"quality")==0) {
   i++; 
   if (i<=argc) {
    JPEGQUALITY = atoi(argv[i]);
    if (JPEGQUALITY<1) JPEGQUALITY=1;
    if (JPEGQUALITY>100) JPEGQUALITY=100;
   }
  } else if (strcmp(arg,"channel")==0) {
   i++; 
   if (i<=argc) {
    INPUTCHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"videodev",2)==0) {
   i++; 
   if (i<=argc) {
    strcpy (VIDEODEV,argv[i]);
   }
  } else if (strncmp(arg,"ntsc",2)==0) {
   VIDEOMODE=VIDEOMODE_NTSC;
  } else if (strncmp(arg,"secam",2)==0) {
   VIDEOMODE=VIDEOMODE_SECAM;
  } else {
   /* Command line help */
   printf ("Command line parameters:\n");
   printf ("\nVideo settings:\n");
   printf (" -width #		width of images (default: %i)\n",WIDTH);
   printf (" -height #		height of images (default: %i)\n",HEIGHT);
   printf (" -channel #		input channel (default: %i)\n",INPUTCHANNEL);
   printf (" -videodev txt	video device (default: %s)\n",VIDEODEV);
   printf (" -ntsc		use NTSC video (default: pal)\n");
   printf (" -secam		use SECAM video (default: pal)\n");
   printf ("\nNetwork settings:\n");
   printf (" -port #		listening port (default: %i)\n",PORT);
   printf (" -maxcon #		maximum simultanous connections (default: %i)\n",CONNECTIONS);
   printf ("\nJPEG settings:\n");
   printf (" -gray #		output grayscale images (default: %i)\n",JPEGGRAY);
   printf (" -quality #		set compression quality (default: %i)\n", JPEGQUALITY);
   exit(0);
  }
 }
  
 /* Message */
 printf ("Videodevice: %s, Channel: %i, Size: %ix%i, Mode: %i.\n",VIDEODEV, INPUTCHANNEL, WIDTH, HEIGHT, VIDEOMODE);
 /* Framegrabber initialization */
 if (fg_open_device (&fg, VIDEODEV)!=0) exit(-1);
 
 if (fg_print_info (&fg)!=0) exit(-1);

 if (fg_set_channel (&fg, INPUTCHANNEL, VIDEOMODE)!=0) exit(-1);

 if (fg_start_grab_image(&fg, WIDTH, HEIGHT, FORMAT_RGB565)!=0) exit(-1);
 printf ("Webcam running ...\n"); 

 /* Network initialization */
 socket_port=PORT;

 /* Setup hostname, port and connection type */
 gethostname(hostname,sizeof(hostname));
 address.sin_family = AF_INET;
 address.sin_port = htons(socket_port);
 address.sin_addr.s_addr = INADDR_ANY;

 /* Create listening socket */
 listening_socket = socket(AF_INET, SOCK_STREAM, 0);
 if (listening_socket < 0) {
  perror("socket");
  exit(-1);
 }
 /* Set options for socket */
 value=1;
 setsockopt(listening_socket,SOL_SOCKET,SO_REUSEADDR,&value,sizeof(value));
 /* Bind socket to address */
 if (bind(listening_socket,(struct  sockaddr *)&address,sizeof(address)) < 0) {
  perror("bind");
  close(listening_socket);
  exit(-1);
 }
 /* Allow multiple connections */
 listen(listening_socket, CONNECTIONS);

 /* Initialize mutex and variables */
 if (init_network_socket(&nt)!=0) exit(-1);

 /* Store framegrabber handle and other info in netthread structure */
 nt.fg=(void *)&fg;
 nt.jpeggray=JPEGGRAY;
 nt.jpegquality=JPEGQUALITY;

 /* Message */
 printf ("Port: %i, Max. Connections: %i.\n\n",PORT, CONNECTIONS);

 /* Get a connection and handle each with a separate thread */
 in_loop=1;
 while (in_loop) {
  length=sizeof(address);
  connected_socket = accept(listening_socket,(struct  sockaddr *)&address, &length);
  if ((connected_socket < 0) && (errno != EINTR)) {
   perror("accept");
   in_loop=0;
  } else {
   /* Store file handle of connected socket and create thread */
   if (start_network_socket(&nt, connected_socket)!=0) {
    in_loop=0;
   }
  }
 }

 /* Shutdown socket connection */
 close(listening_socket);

 /* Stop framegrabber */
 if (fg_stop_grab_image(&fg)!=0) exit(-1); 	
 if (fg_close_device (&fg)!=0) exit(-1); 	
 
 exit(0);
}
