/*
    Support: Threaded socket routines for webcam
    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer    -    Deskplayer Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |
*/

#include "bgrab.h"
#include "webcamsocket.h"

/* Converts ascii text to in_addr struct */

struct in_addr *a_to_addr(char *address)
{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if (saddr.s_addr != -1) {
    return &saddr;
  }
    
  host = gethostbyname(address);
  if (host != NULL) {
    return (struct in_addr *) *host->h_addr_list;
  }
  return NULL;
}

/* Socket IO */
   
int sock_write (int sockfd, char *buf, size_t count)
{
 size_t bytes_sent = 0;
 int this_write;
       
 while (bytes_sent < count) {
  do {
   this_write = write(sockfd, buf, count - bytes_sent);
  } while ( (this_write < 0) && (errno == EINTR) );
  if (this_write <= 0) {
   return(this_write);
  } 
  bytes_sent += this_write;
  buf += this_write;
 }
 return(count);
}

int sock_gets (int sockfd, char *str, size_t count)
{
 int bytes_read;
 int total_count = 0;
 char *current_position;
 char last_read = 0;
 fd_set readers;
 struct timeval timeout;
 int pending;
 int polltimeout;
          
 current_position = str;
 while ((last_read != 10) && (last_read !=13)) {
   /* Check status of connection */
   pending=0;
   polltimeout = 50;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=0;
    timeout.tv_usec=100000;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if ((polltimeout>0) && (pending<0)) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
    polltimeout--;
   }
   if (polltimeout==0) {
    return(total_count);
   } 
   bytes_read = read(sockfd, &last_read, 1);
   if ((bytes_read==0) && (pending==1)) {
    fprintf (stderr,"Connection closed ...\n");
    return(-1);
   }    
   if (bytes_read <= 0) {
    return bytes_read; 
   }
   if ( (total_count < count) && (last_read != 10) && (last_read !=13) ) {
    *current_position = last_read;
    current_position++;
    total_count++;
   }
 }
 if (count > 0) *current_position = 0;
 return total_count;
}

/* Network service thread                 */

int NetworkThread(struct netthreads *nt)
{
 int connected_socket;
 struct fgdevice *fg;
 char *newimage;
 char inbuffer[BUFFER_LEN];
 char outbuffer[BUFFER_LEN];
 int connected, outlen;
 FILE *tempfile;
 FILE *jpegpipe; 
 char commandname[LINE_LEN];
 char tempfilename[LINE_LEN];
 char errormessage[LINE_LEN]; 
 int rval,gval,bval;
 int rowlength;
 unsigned char *rgbtriples, *curtriple;  
 unsigned short *curpos; 
 int x,y;
  
 /* Transfer socket and framegrabber handle to local variable */
 shared_data_lock(nt);
 connected_socket=nt->connected_socket;
 fg=(struct fgdevice *)nt->fg; 
 shared_data_unlock(nt);
 /* Create local image and conversion buffer */
 if ((newimage=(char *)malloc(fg->image_size))==NULL) {
  fprintf (stderr,"Out of memory.");
  return(-1);
 }
 rowlength=3*fg->width;
 if ((rgbtriples=(unsigned char *)malloc(rowlength))==NULL) {
  fprintf (stderr,"Out of memory.");
  return(-1);
 }
 /* Go into processing loop */
 connected=1;
 while (connected) {
   /* Read mandatory GET line */
   if (sock_gets(connected_socket,inbuffer,BUFFER_LEN)==0) {
    connected=0;
   } else {
    /* Get an image with locking at netthread level */
    shared_data_lock(nt);
    /* Update framebuffer with new image */
    if (fg_get_next_image(fg)!=NULL) {
     /* Copy image to local buffer */
     memcpy(newimage,(char *)fg->current_image,fg->image_size);
    } else {
     /* Blank local buffer */
     memset(newimage,0,fg->image_size);
    }
    shared_data_unlock(nt);
    /* Save .ppm image as tempfile */
    shared_data_lock(nt);
    sprintf (tempfilename,"/tmp/webcam.%i",nt->camimage++);
    shared_data_unlock(nt);
    if (strlen(tempfilename)>1) {
     if ( (tempfile=fopen(tempfilename,"w"))!=NULL) {
      /* Write header */
      fprintf(tempfile,"P6\n# Webcam\n%i %i\n255\n",fg->width,fg->height);
      /* Dump screen into pipe*/
      curpos=(unsigned short *)newimage;
      for (y=0; y<fg->height; y++) {
       curtriple=rgbtriples;
       for (x=0; x<fg->width; x++) {
        rval=(*curpos & 0xf800); rval = rval >> 8;
        gval=(*curpos & 0x07d0); gval = gval >> 3;
        bval=(*curpos & 0x001f); bval = bval << 3;
        *curtriple=(unsigned char)rval;
        curtriple++;
        *curtriple=(unsigned char)gval;
        curtriple++;
        *curtriple=(unsigned char)bval;
        curtriple++;
        curpos++;
       }
       fwrite ((void *)rgbtriples,rowlength,1,tempfile);
      }
      fclose(tempfile);
     }
     /* Convert data to jpeg via an external program */
     /* Open commandpipe */
     sprintf(commandname,"/usr/bin/cjpeg -quality %i ",nt->jpegquality);
     if (nt->jpeggray) {
      strcat(commandname,"-grayscale ");
     }
     strcat(commandname,tempfilename);
     if ((jpegpipe=popen(commandname,"r"))==NULL) {
      sprintf(errormessage,"popen(%s)",commandname);
      perror(errormessage);
      connected=0;
     } else { 
      /* Transfer HTML header and jpeg-data out to socket */           
      if (connected) {
       /* Header lines */
       sprintf (outbuffer,"HTTP/1.1 200 OK\nContent-Type: image/jpeg\n\n");
       /* Send data to network socket */
       connected=(sock_write(connected_socket,outbuffer,strlen(outbuffer))>0);
      }
      while (connected) {
       /* Read from pipe */
       outlen=fread(outbuffer,1,BUFFER_LEN,jpegpipe);
       /* Send data to network socket */
       connected=((outlen=sock_write(connected_socket,outbuffer,outlen))>0);
       /* Determine EOF condition */
       if ((outlen<BUFFER_LEN) || feof(jpegpipe)) {
        connected=0;
       }
      }
      /* Close pipe */
      pclose(jpegpipe);
     }
    } 
    /* Remove tempfilename */
    remove (tempfilename);
   }
 }
 /* Clean up */
 if (newimage!=NULL) free(newimage);
 /* Flush data */
 fflush(NULL);
 /* Read client messages */
 recv(connected_socket,outbuffer,BUFFER_LEN,0);
 /* Exit and goodby */
 close(connected_socket);
 return(0);
}

int shared_data_lock(struct netthreads *nt)
{
 /* Lock shared variables */ 
 pthread_mutex_lock(&nt->network_mutex);
 
 return(0);
}

int shared_data_unlock(struct netthreads *nt)
{
 /* Unlock shared variables */ 
 pthread_mutex_unlock(&nt->network_mutex);
 
 return(0);
}

int start_network_socket(struct netthreads *nt, int connected_socket)
{
 pthread_t network_thread;

 /* Store socket handle */
 shared_data_lock(nt);
 nt->connected_socket=connected_socket;
 shared_data_unlock(nt);
 /* Start thread */
 if (pthread_create(&network_thread, NULL,(void *)NetworkThread,(void *)nt)==-1) {
  perror ("pthread_create()");
  return(-1);
 } else {
  return(0);
 } 
}

int init_network_socket(struct netthreads *nt)
{
 /* Initialize sycro mutex */
 if (pthread_mutex_init(&nt->network_mutex, NULL)==-1) {
  perror("pthread_mutex_init()");
  return(-1);
 }
 nt->camimage=0;
 
 return(0);
}
