/*

    Subroutine: Utility routines using XWindows access through shared memory.

    The contents of this file are subject to the Mozilla Public
    License Version 1.1 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/
    
    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.
   
    The Original Code is the libbgrab framegrabber library.
    
    The Initial Developer of the Original Code is A. Schiffler.  
    Portions created by A. Schiffler are Copyright (C) 1999-2000
    A. Schiffler.  All Rights Reserved.
    
    Contributor(s): Andreas Schiffler <aschiffler@home.com>
                    M. Hoch <mpunkt@zkm.de>
                    Peter Serocka <pserocka@Mathematik.Uni-Bielefeld.de>
                                                                                                                                  dnl 
    Alternatively, the contents of this file may be used under the
    terms of the GNU General Public License Version 2 or later (the
    "GPL"), in which case the provisions of the GPL are applicable 
    instead of those above.  If you wish to allow use of your 
    version of this file only under the terms of the GPL and not to
    allow others to use your version of this file under the MPL,
    indicate your decision by deleting the provisions above and
    replace them with the notice and other provisions required by
    the GPL.  If you do not delete the provisions above, a recipient
    may use your version of this file under either the MPL or the
    GPL.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  Senior Systems Engineer  -  Deskplayer.Com Inc., Buffalo  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/

/* Additional colormap support by M. Hoch */
/* Proper XVisuals selection support by Peter Serocka */

#include "xutil.h"

#include "font8x8.h"

int setup_xwinbuffer (struct xwinbuffer *xwin, const char *prgname, int width, int height, int depth_in_byte)
{
 int depth_in_bit;
 int screen;
 XVisualInfo desired_visual_template;
 int n_visuals_matched;
 XVisualInfo *visual_info_list;
 unsigned int attrib_mask;
 XSetWindowAttributes attribs;
 Visual *visual;
 Colormap wtcolormap;
 int numCols;
 XColor *palette;
 int pixelNum;
 XGCValues gc_values;

 /* Store settings in structure */
 xwin->width = width;
 xwin->height = height;
 xwin->depth = depth_in_byte;
 
 /* Init variables */
 depth_in_bit = 8 * depth_in_byte;

 /* Reset mode-switch variables */ 
 xwin->modeswitch=0;
 xwin->vm_smode=-1;

 /* Calculate bounds */
 xwin->maxx=width-1;
 xwin->maxy=height-1;

 /* 1. Connect to the X Server */
 xwin->display_ = XOpenDisplay (NULL);
 
 /* XSynchronize (xwin->display_, 1); */
 screen = DefaultScreen (xwin->display_);

 /* 2. Find an X visual */
 if (depth_in_bit == 8) {
  desired_visual_template.class = PseudoColor;
 } else {
  desired_visual_template.class = TrueColor;
 }
 desired_visual_template.depth = depth_in_bit;
 desired_visual_template.screen = screen;
 /* ... look */
 n_visuals_matched = 0;
 visual_info_list = XGetVisualInfo (xwin->display_,
            VisualScreenMask | VisualClassMask | VisualDepthMask,
            &desired_visual_template,
            &n_visuals_matched);
 if (n_visuals_matched == 0) {
  fprintf (stderr, "No visual matched.\n");
  return (-1);
 }
 // Take the first visual:
 visual = visual_info_list[0].visual;

 /* 3. Define attributes for the window */
 attrib_mask = CWBackPixel | CWBorderPixel;
 attribs.background_pixel = BlackPixel (xwin->display_, screen);
 attribs.border_pixel = BlackPixel (xwin->display_, screen);

 // 4. Create a Colormap
 if (desired_visual_template.class == PseudoColor) {
  // color map needed when using BW-Images
  fprintf (stderr, "Creating colormap.\n");
  wtcolormap = XCreateColormap (xwin->display_,
                DefaultRootWindow (xwin->display_),
                visual_info_list[0].visual, AllocAll);

  numCols = 1 << depth_in_bit;
  palette = (XColor *) malloc (numCols * sizeof (XColor));

  for (pixelNum = 0; pixelNum < numCols; pixelNum++) {
            palette[pixelNum].pixel = pixelNum;
            palette[pixelNum].red = pixelNum << (16 - depth_in_bit);
            palette[pixelNum].green = pixelNum << (16 - depth_in_bit);
            palette[pixelNum].blue = pixelNum << (16 - depth_in_bit);
            palette[pixelNum].flags = DoRed | DoGreen | DoBlue;
  }

  XStoreColors (xwin->display_, wtcolormap, palette, numCols);

  attrib_mask |= CWColormap;
  attribs.colormap = wtcolormap;

 }

 // 3. Create the Window
 xwin->wtwin_ = XCreateWindow (xwin->display_, DefaultRootWindow (xwin->display_),
            0, 0, xwin->width, xwin->height, 0,
            depth_in_bit, InputOutput, visual, attrib_mask, &attribs);
  XStoreName (xwin->display_, xwin->wtwin_, prgname);
  XMapWindow (xwin->display_, xwin->wtwin_);
  XSelectInput (xwin->display_, xwin->wtwin_, KeyPressMask | KeyReleaseMask | ButtonPressMask);

 // 4. create a graphics context
 gc_values.graphics_exposures = False;
 xwin->gc_ = XCreateGC (xwin->display_, xwin->wtwin_, GCGraphicsExposures, &gc_values);

 /* 5. Create shared memory segment for direct framebuffer access */
 xwin->shminfo = (XShmSegmentInfo *) malloc (sizeof (XShmSegmentInfo));
 xwin->shmimage = XShmCreateImage (xwin->display_, visual,
        depth_in_bit, ZPixmap, NULL, xwin->shminfo,
        xwin->width, xwin->height);

 /* Print info */
 fprintf (stderr,
        "Xwindows: depth (bits): %d, size: "
        "(%i, %i); shm-bpl: %i, shm-height: %i, shm-size: %i\n",
        depth_in_bit, xwin->width, xwin->height,
        xwin->shmimage->bytes_per_line, xwin->shmimage->height,
        xwin->shmimage->bytes_per_line * xwin->shmimage->height);

 /* Attach shm segment */
 if ((xwin->shminfo->shmid = shmget (getpid (),
                xwin->shmimage->bytes_per_line * xwin->shmimage->height,
                IPC_CREAT | IPC_EXCL | 0777)) == -1) {
        perror ("shmget()");
        return (-1);
 }
 xwin->shminfo->shmaddr =
        (char *) shmat (xwin->shminfo->shmid, (void *) 0, 0);
 xwin->shmimage->data = xwin->shminfo->shmaddr;
 XShmAttach (xwin->display_, xwin->shminfo);

 return 0;
}

/* Write shared memory content with display content */

int sync_xwinbuffer(struct xwinbuffer *xwin)
{
 /* Draw screen onto display */
 XShmPutImage(xwin->display_, xwin->wtwin_, xwin->gc_, xwin->shmimage,
 	        0, 0, 0, 0, xwin->width, xwin->height, False);
 XSync(xwin->display_, 0); 
 return(0);
}

/* Read keyboard and mouse events */

int keypoll_xwinbuffer (struct xwinbuffer *xwin, KeySym *key, int *button, int *mousex, int *mousey)
{
  XEvent event; 
 
  /* Check keyboard/mouse input */
  if (XPending(xwin->display_)>0) {
   *button=0;
   *key=0;
   XNextEvent(xwin->display_, &event);
   switch (event.type) {
    case ButtonPress:
     *mousex=event.xbutton.x;
     *mousey=event.xbutton.y;
     *button=1;
     break;
    case KeyPress:
     *key=XLookupKeysym((void *)&event,0);
     break;
    case MappingNotify:
     XRefreshKeyboardMapping((void  *)&event);
     break;
   }
   return(1);
  } else { 
   return(0);
  } 
}  

/* Close the window */

int close_xwinbuffer (struct xwinbuffer *xwin)
{
 /* Reset video-mode if it was changed */
 if (xwin->modeswitch) {
  videoModeReset(xwin);
 }

 /* Mark Shm as destroyed */
 shmctl(xwin->shminfo->shmid,IPC_RMID,NULL);

 /* Detatch shared memory segment */
 shmdt ((char *)xwin->shminfo->shmaddr);
 
 return(0);
}

/* Drawing utilities */

/* Clip X to screen coordinates */

void ClipX(struct xwinbuffer *xwin, int *x)
{
 /* Clip X coordinates */
 if (*x<0) {
  *x=0;
 } else if (*x>xwin->maxx) {
  *x=xwin->maxx;
 } 
}

/* Clip Y to screen coordinates */

void ClipY(struct xwinbuffer *xwin, int *y)
{
 /* Clip Y coordinates */
 if (*y<0) {
  *y=0;
 } else if (*y>xwin->maxy) {
  *y=xwin->maxy;
 } 
}

/* Draw horizontal line into color buffer */

void hline (struct xwinbuffer *xwin, int x1, int x2, int y, unsigned int color)
{
 int i;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 unsigned int *curpos32,color32;
 
 if (xwin==NULL) return; 
 /* Clip coordinates */
 ClipX(xwin,&x1);
 ClipX(xwin,&x2);
 ClipY(xwin,&y);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x1+y*xwin->width);
  for (i=0; i<(x2-x1+1); i++) {
   *curpos8=color8;
   curpos8++;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x1+y*xwin->width);
  for (i=0; i<=(x2-x1); i++) {
   *curpos16=color16;
   curpos16++;
  }
 } else if (xwin->depth==4) {
  color32=(unsigned int)color;
  curpos32 = (unsigned int *)xwin->shmimage->data;
  curpos32 += (x1+y*xwin->width);
  for (i=0; i<=(x2-x1); i++) {
   *curpos32=color32;
   curpos32++;
  }
 }
}

/* Draw vertical line into color buffer */
void vline (struct xwinbuffer *xwin, int x, int y1, int y2, unsigned int color)
{
 int i;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 unsigned int *curpos32,color32;

 if (xwin==NULL) return; 
 /* Clip coordinates */
 ClipX(xwin,&x);
 ClipY(xwin,&y1);
 ClipY(xwin,&y2);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos8=color8;
   curpos8 += xwin->width;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos16=color16;
   curpos16 += xwin->width;
  }
 } else if (xwin->depth==4) {
  color32=(unsigned int)color;
  curpos32 = (unsigned int *)xwin->shmimage->data;
  curpos32 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos32=color32;
   curpos32 += xwin->width;
  }
 }
}

/* Draw character into buffer */
void drawchar (struct xwinbuffer *xwin, int x, int y, const char c, unsigned int color)
{
 int i,j,k;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 unsigned int *curpos32,color32;
 unsigned char *charpos;
 unsigned char bits[8]={128,64,32,16,8,4,2,1};
 unsigned char *bitpos;
 int ystep,xmax,ymax;

 if (xwin==NULL) return; 
 /* Clip coordinates */
 xmax=x+8;
 ymax=y+8;
 ClipX(xwin,&x);
 ClipY(xwin,&y);
 ClipX(xwin,&xmax);
 ClipY(xwin,&ymax);
 /* Variable setup */
 k = (unsigned char)c;
 k *= 8;
 charpos=font8x8;
 charpos += k;
 bitpos=bits;
 ystep=xwin->width-(xmax-x);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y*xwin->width);
  for (i=y; i<ymax; i++) {
   bitpos=bits;
   for (j=x; j<xmax; j++) {
    if ((*charpos & *bitpos)==*bitpos) {
     *curpos8=color8;
    } 
    curpos8++;
    bitpos++;
   }
   curpos8 += ystep;
   charpos++;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y*xwin->width);
  for (i=y; i<ymax; i++) {
   bitpos=bits;
   for (j=x; j<xmax; j++) {
    if ((*charpos & *bitpos)==*bitpos) {
     *curpos16=color16;
    } 
    curpos16++;
    bitpos++;
   }
   curpos16 += ystep;
   charpos++;
  }
 } else if (xwin->depth==4) {
  color32=(unsigned int)color;
  curpos32 = (unsigned int *)xwin->shmimage->data;
  curpos32 += (x+y*xwin->width);
  for (i=y; i<ymax; i++) {
   bitpos=bits;
   for (j=x; j<xmax; j++) {
    if ((*charpos & *bitpos)==*bitpos) {
     *curpos32=color32;
    } 
    curpos32++;
    bitpos++;
   }
   curpos32 += ystep;
   charpos++;
  }
 }
}

/* Draw string into buffer */
void drawstring (struct xwinbuffer *xwin, int x, int y, const char *s, unsigned int color)
{
 int i,length;
 const char *curchar;
 int curx;

 length=strlen(s);
 curchar=s;
 curx=x;
 for (i=0; i<length; i++) {
  drawchar(xwin,curx,y,*curchar,color);
  curchar++;
  curx += 8;
 } 
}

/* Draw Box into color buffer */
void drawbox (struct xwinbuffer *xwin, int x1, int y1, int x2, int y2, unsigned int color)
{
 int i,j,wstep;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 unsigned int *curpos32,color32;

 if (xwin==NULL) return; 
 /* Clip coordinates */
 ClipX(xwin,&x1);
 ClipX(xwin,&x2);
 ClipY(xwin,&y1);
 ClipY(xwin,&y2);
 /* Width step */
 wstep=(xwin->width-(x2-x1+1));
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x1+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   for (j=x1; j<=x2; j++) {
    *curpos8=color8;
    curpos8++;
   }
   curpos8 += wstep;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x1+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   for (j=x1; j<=x2; j++) {
    *curpos16=color16;
    curpos16++;
   }
   curpos16 += wstep;
  }
 } else if (xwin->depth==4) {
  color32=(unsigned int)color;
  curpos32 = (unsigned int *)xwin->shmimage->data;
  curpos32 += (x1+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   for (j=x1; j<=x2; j++) {
    *curpos32=color32;
    curpos32++;
   }
   curpos32 += wstep;
  }
 }
}

void drawpixel(struct xwinbuffer *xwin, int x, int y, unsigned int color)
{
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 unsigned int *curpos32,color32;

 /* Clip coordinates */
 ClipX(xwin,&x);
 ClipY(xwin,&y);
 /* Draw */
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y*xwin->width);
  *curpos8=color8;
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y*xwin->width);
  *curpos16=color16;
 } else if (xwin->depth==4) {
  color32=(unsigned int)color;
  curpos32 = (unsigned int *)xwin->shmimage->data;
  curpos32 += (x+y*xwin->width);
  *curpos32=color32;
 }
}

/* Draw pixel into 8bpp screen without clipping */
void drawpixel8fast(struct xwinbuffer *xwin, int x, int y, unsigned int color)
{
 unsigned char *curpos8;

 curpos8 = (unsigned char *)xwin->shmimage->data;
 curpos8 += (x+y*xwin->width);
 *curpos8=(unsigned char)color;
}

/* Draw pixel into 16bpp screen without clipping */
void drawpixel16fast(struct xwinbuffer *xwin, int x, int y, unsigned int color)
{
 unsigned short *curpos16;
 
 curpos16 = (unsigned short *)xwin->shmimage->data;
 curpos16 += (x+y*xwin->width);
 *curpos16=(unsigned short)color;
}

/* Draw pixel into 32bpp screen without clipping */
void drawpixel32fast(struct xwinbuffer *xwin, int x, int y, unsigned int color)
{
 unsigned int *curpos32;
 
 curpos32 = (unsigned int *)xwin->shmimage->data;
 curpos32 += (x+y*xwin->width);
 *curpos32=(unsigned int)color;
}

#define ABS(a) (((a)<0) ? -(a) : (a))

/* Draw a line between two points */

void drawline(struct xwinbuffer *xwin, int x1, int y1, int x2, int y2, unsigned int color)
{
 int dx;
 int dy;
 int ax;
 int ay;
 int sx;
 int sy;

 /* Variable setup and clipping */
 int x = x1;
 int y = y1;
 ClipX(xwin,&x1);
 ClipY(xwin,&y1);
 ClipX(xwin,&x2);
 ClipY(xwin,&y2);
 dx = x2 - x1;
 dy = y2 - y1;
 ax = ABS(dx) << 1;
 ay = ABS(dy) << 1;
 sx = (dx >= 0) ? 1 : -1;
 sy = (dy >= 0) ? 1 : -1;
 x = x1;
 y = y1;
 /* Depth dependent drawing routines */
 if (xwin->depth==1) {
  if (ax > ay) {
   int d = ay - (ax >> 1);
   while (x != x2) {
    drawpixel8fast(xwin, x, y, color);
    if (d > 0 || (d == 0 && sx == 1)) {
     y += sy;
     d -= ax;
    } 
    x += sx;
    d += ay;
   }
  } else {
   int d = ax - (ay >> 1);
   while (y != y2) {
    drawpixel8fast(xwin, x, y, color);
    if (d > 0 || (d == 0 && sy == 1)) {
     x += sx;
     d -= ay;
    }
    y += sy;
    d += ax;
   }
  }
  drawpixel8fast(xwin, x, y, color);
 } else if (xwin->depth==2) {
  if (ax > ay) {
   int d = ay - (ax >> 1);
   while (x != x2) {
    drawpixel16fast (xwin, x, y, color);
    if (d > 0 || (d == 0 && sx == 1)) {
     y += sy;
     d -= ax;
    }
    x += sx;
    d += ay;
   }
  } else {
   int d = ax - (ay >> 1);
   while (y != y2) {
    drawpixel16fast (xwin, x, y, color);
    if (d > 0 || (d == 0 && sy == 1)) {
     x += sx;
     d -= ay;
    }
    y += sy;
    d += ax;
   }
  }
  drawpixel16fast (xwin, x, y, color);
 } else if (xwin->depth==4) {
  if (ax > ay) {
   int d = ay - (ax >> 1);
   while (x != x2) {
    drawpixel32fast (xwin, x, y, color);
    if (d > 0 || (d == 0 && sx == 1)) {
     y += sy;
     d -= ax;
    }
    x += sx;
    d += ay;
   }
  } else {
   int d = ax - (ay >> 1);
   while (y != y2) {
    drawpixel32fast (xwin, x, y, color);
    if (d > 0 || (d == 0 && sy == 1)) {
     x += sx;
     d -= ay;
    }
    y += sy;
    d += ax;
   }
  }
  drawpixel32fast (xwin, x, y, color);
 } // depth check 
}

/* Video mode change routines */

/* Change the current X11 video mode */

void videoModeChange(struct xwinbuffer *xwin, int width, int height)
{
 int i;

 XF86VidModeGetAllModeLines(xwin->display_,XDefaultScreen(xwin->display_), &xwin->vm_count,&xwin->vm_modelines);
 for (i = 0; i < xwin->vm_count; i++) {
  if (xwin->vm_modelines[i]->hdisplay == width && xwin->vm_modelines[i]->vdisplay == height) {
   xwin->vm_smode=i;
  }
 }

 if (xwin->vm_smode != -1) {
  XF86VidModeSwitchToMode(xwin->display_,XDefaultScreen(xwin->display_),xwin->vm_modelines[xwin->vm_smode]);
  XFlush(xwin->display_);
  xwin->modeswitch=1;
 }
}

/* Reset X11 video mode */

void videoModeReset(struct xwinbuffer *xwin)
{
 if (xwin->vm_smode != -1) {
  XF86VidModeSwitchToMode(xwin->display_,XDefaultScreen(xwin->display_),xwin->vm_modelines[0]);
  XFlush(xwin->display_);
 }
 xwin->vm_smode=-1;
 xwin->modeswitch=0;
}
